import numpy as np
import torch
import matplotlib.pyplot as plt
import geopandas as gpd
import os
import json
import glob
import tmp as utils
import pandas as pd

def iou_numpy(outputs, labels):
    """
    Calculate the Intersection over Union (IoU) score between binary segmentation outputs and ground truth labels.

    Args:
        outputs (torch.Tensor): Binary segmentation outputs.
        labels (torch.Tensor): Ground truth labels.

    Returns:
        torch.Tensor: IoU score.
    """
    intersection = torch.logical_and(labels, outputs)
    union = torch.logical_or(labels, outputs)
    iou_score = torch.sum(intersection) / torch.sum(union)
    return iou_score


class MatchingAlgorithm():
    """
    Initialize the MatchingAlgorithm.
    This class is designed to perform matching between ground truth bounding boxes
    and predicted bounding boxes based on Intersection over Union (IoU) scores.
    It takes lists of ground truth (GT) and predicted bounding boxes along with
    an optional IoU threshold for matching.

    Args:
        gt_bbox (List[np.ndarray]): List of ground truth bounding boxes.
        pred_bbox (List[np.ndarray]): List of predicted bounding boxes.
        iou_threshold (float): Intersection over Union (IoU) threshold for matching. Default is 0.5.
    """
    def __init__(self, gt_bbox, pred_bbox, iou_threshold=0.5):
        self.gt_bboxes = gt_bbox
        self.pred_bboxes = pred_bbox
        self.iou_threshold = iou_threshold

    def matching(self):
        """ 
        This method calculates the IoU scores between all pairs of ground truth
        and predicted bounding boxes. It then identifies true positives, false positives,
        and false negatives based on the IoU threshold. The method returns various metrics
        including IoU list, F1 scores, indices of true positives, false positives, and false negatives.

        Returns:
            Tuple of results including IoU list, F1 scores, indices of true positives,
            indices of false positives, indices of false negatives, matching scores,
            precision, and recall.
        """
        if len(self.pred_bboxes) == 0 or len(self.gt_bboxes) == 0:
            print("Both predicted and ground truth bounding boxes are empty.")
            return [], [], [], [], [], [], [], [], []

        iou_matrix = np.zeros((len(self.pred_bboxes), len(self.gt_bboxes)))

        for i in range(len(self.pred_bboxes)):
            for j in range(len(self.gt_bboxes)):
                iou_matrix[i, j] = iou_numpy(torch.from_numpy(self.pred_bboxes[i]), torch.from_numpy(self.gt_bboxes[j]))

        iou_list = []
        f1_scores = []
        pred_matched = set()
        gt_matched = set()
        tp_pred_indices = []
        tp_gt_indices = []
        m_score=[]
        mscores=[]

        while True:
            max_iou = np.max(iou_matrix)
            if max_iou < self.iou_threshold:
                break
            max_index = np.unravel_index(
                np.argmax(iou_matrix), iou_matrix.shape)
            iou_list.append(max_iou)
            pred_matched.add(max_index[0])
            gt_matched.add(max_index[1])

            tp_pred_indices.append(max_index[0])
            tp_gt_indices.append(max_index[1])

            f1_score = 2 * max_iou / (max_iou + 1)
            f1_scores.append(f1_score)

            print(
                f"Matched predicted box {max_index[0]} with GT box {max_index[1]}, IoU = {max_iou}, F1 = {f1_score}")
            m_score={
                'pred_box':int(max_index[0]),
                'GT_box':int(max_index[1]),
                'iou':float(max_iou),
                'f1':float(f1_score)
            }
            mscores.append(m_score)
            iou_matrix[max_index[0], :] = 0
            iou_matrix[:, max_index[1]] = 0

        for i in set(range(len(self.pred_bboxes))) - pred_matched:
            iou_list.append(0)
            f1_scores.append(0)
            print(f"Unmatched predicted box {i} has no match, IoU = 0, F1 = 0")

        for i in set(range(len(self.gt_bboxes))) - gt_matched:
            iou_list.append(0)
            f1_scores.append(0)
            print(f"Unmatched GT box {i} has no match, IoU = 0, F1 = 0")

        print("number of GT boxes:", len(self.gt_bboxes))
        print("number of predicted boxes:", len(self.pred_bboxes))

        fp_indices = list(set(range(len(self.pred_bboxes))) - pred_matched)
        fn_indices = list(set(range(len(self.gt_bboxes))) - gt_matched)

        true_positives = len(tp_pred_indices)
        false_positives = len(fp_indices)
        false_negatives = len(fn_indices)

        precision = true_positives / (true_positives + false_positives)
        recall = true_positives / (true_positives + false_negatives)

        return iou_list, f1_scores, tp_pred_indices, tp_gt_indices, fp_indices, fn_indices, mscores, precision, recall

    def tp_iou(self, tp_pred_indices, tp_gt_indices):
        """
        Calculate IoU for true positive matches by taking the indices of true positive predicted and 
        ground truth boxes and calculating the Intersection over Union (IoU) scores for each true positive match.
        It returns a tuple containing the list of IoU scores for true positive matches
        and the average IoU.

        Args:
            tp_pred_indices (List[int]): Indices of true positive predicted boxes.
            tp_gt_indices (List[int]): Indices of true positive ground truth boxes.

        Returns:
            Tuple containing the list of IoU scores for true positive matches and the average IoU.
        """
        tp_iou_list = []
        for i, j in zip(tp_pred_indices, tp_gt_indices):
            iou = iou_numpy(torch.from_numpy(self.pred_bboxes[i]), torch.from_numpy(self.gt_bboxes[j]))
            tp_iou_list.append(float(np.nan_to_num(iou)))

        if len(tp_iou_list) > 0:
            avg_tp_iou = float(np.mean(tp_iou_list))
        else:
            avg_tp_iou = None
        return tp_iou_list, avg_tp_iou


class CalScores:
    """
    Initialize the CalScores class.
    Designed for evaluating segmentation results, particularly for building detection tasks. It provides 
    methods to calculate both micro-level and macro-level matching metrics between predicted and 
    ground truth building masks. Additionally, it facilitates the storage and visualization of results, 
    such as saving score files and generating annotated images.

    Args:
        output_dir (str): Directory to save output files.
        score_dir (str): Directory to save score files.
    """
    def __init__(self,output_dir,score_dir):
        self.output_dir=output_dir
        self.score_dir=score_dir

    def micro_match_iou(self,pred_mask, name, gt, image,input_point,input_label,tile_boxes, save=None, visualize=None):
        """
        Calculate micro-level matching metrics and save results.
        Responsible for calculating micro-level matching metrics between predicted and ground truth 
        building masks for a specific image tile. It takes into account various metrics such as IoU 
        (Intersection over Union), F1 score, precision, and recall. The results are stored in a JSON file, 
        and an annotated image is generated for visual inspection.

        Args:
            pred_mask (torch.Tensor): Predicted mask.
            name (str): Name identifier for the image.
            gt (dict): Ground truth information.
            image (numpy.ndarray): Input image.
            input_point: Input points.
            input_label: Input labels.
            tile_boxes: Boxes related to tiles.
            save: Flag set for saving.
            visualize: Flag set for visualization.
        """
        pred_tile = []
        gt_tile = []
        msk = pred_mask.int()
        msk = msk.cpu().numpy()
        scores_b = []
        score = {}
        mask_tile = np.zeros(image.shape[:2])

        for i in range(msk.shape[0]):
            batch = msk[i]
            for b in range(batch.shape[0]):
                mask_tile = mask_tile + batch[b]
                pred_tile.append(batch[b])


        gt_tile=utils.convert_polygon_to_mask_batch(gt['geometry'])

        matcher = MatchingAlgorithm(gt_tile, pred_tile)
        iou_list, f1_scores, tp_pred_indices, tp_gt_indices, fp_indices, fn_indices, mscores, precision,recall = matcher.matching()
        tp_iou_list, avg_tp_iou = matcher.tp_iou(tp_pred_indices, tp_gt_indices)

        score['iou_list'] = iou_list
        score['f1_scores'] = f1_scores
        score['tp_iou_list'] = tp_iou_list
        score['fp_indices'] = fp_indices
        score['fn_indices'] = fn_indices
        score['Mean_iou'] = np.mean(iou_list, dtype=float)
        score['Mean_f1'] = np.mean(f1_scores, dtype=float)
        score['avg_tp_iou'] = float(avg_tp_iou) if avg_tp_iou is not None else 0.0
        score['precision'] = precision
        score['recall'] = recall

        for s in mscores:
            scores_b.append(s)
        scores_b.append(score)

        with open(self.score_dir + f'/{name}_score.json', 'w') as f1:
            json.dump(scores_b, f1)
            
        if save is not None:
            utils.save_shp(pred_mask,name,self.output_dir,image.shape[:2])
        
        plt.figure(figsize=(10, 10))
        plt.imshow(image)
        utils.show_mask(mask_tile, plt.gca(), random_color=False)
        if input_point is not None:
            utils.show_points(input_point.cpu(), input_label.cpu(), plt.gca())
        for box in tile_boxes:
            utils.show_box(box,plt.gca())
        
        if visualize is not None:
            for box in tile_boxes:
                x = []
                y = []
                for i in range(len(box)):
                    if i % 2 == 0:
                        x.append(box[i])
                    else:
                        y.append(box[i])
                plt.plot(x, y)
        plt.show()

        if visualize is not None:
            gtmask=np.zeros((384,384))
            for g in gt_tile:
                gtmask=g+gtmask
            plt.imshow(gtmask)
            plt.show()
        

    def macro_score(self):
        """
        Calculate macro-level matching metrics and save results.
        Aggregates and analyzes the macro-level matching metrics for all image tiles. It reads previously 
        generated score files, calculates average scores, and generates a CSV file summarizing the overall performance. 
        This method provides insights into the global performance of the segmentation model across multiple tiles.
        """
        score_list = []
        all_iou_scores = []
        all_f1_scores = []
        all_precision=[]
        all_recall=[]
        for i in glob.glob(os.path.join(self.score_dir, "*.json")):
            name = i.split("/")[-1]
            name = name.split("_score")[0]

            f = open(i)
            file_data = json.load(f)
            ds = {}
            iou = file_data[len(file_data) - 1]["Mean_iou"]
            f1 = file_data[len(file_data) - 1]["Mean_f1"]
            avg_tp_iou = file_data[len(file_data) - 1]["avg_tp_iou"]
            precision = file_data[len(file_data) - 1]["precision"]
            recall = file_data[len(file_data) - 1]["recall"]

            all_precision.append(precision)
            all_recall.append(recall)

            all_iou_scores.append(file_data[len(file_data) - 1]["iou_list"])
            all_f1_scores.append(file_data[len(file_data) - 1]["f1_scores"])


            ds["name"] = name
            ds["iou"] = iou
            ds["f1"] = f1
            ds["avg_tp_iou"] = avg_tp_iou
            ds["precision"] = precision
            ds["recall"] = recall
            score_list.append(ds)

        df = pd.DataFrame(score_list)
        df.to_csv(self.score_dir + "/scores.csv", index=False)

        all_i = []
        all_f = []
        all_tpi = []
        all_tpf = []

        for i1, f11 in zip(all_iou_scores, all_f1_scores):
            for i2, f12 in zip(i1, f11):
                all_i.append(i2)
                all_f.append(f12)
                if i2 > 0 and f12 > 0:
                    all_tpi.append(i2)
                    all_tpf.append(f12)

        total_iou = np.nanmean(np.array(all_i))
        total_f1 = np.nanmean(np.array(all_f))
        total_tpiou = np.mean(np.array(all_tpi))
        total_tpf1 = np.mean(np.array(all_tpf))
        total_precision = np.mean(np.array(all_precision))
        total_recall = np.mean(np.array(all_recall))

        print("Mean iou score of all buildings in all tiles:", total_iou)
        print("Mean F1 score of all buildings in all tiles:", total_f1)
        print("Mean tp iou score of all buildings in all tiles:", total_tpiou)
        print("Mean tp f1 score of all buildings in all tiles:", total_tpf1)
 



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Mar 26 20:50:11 2021

@author: hasan
"""
#configs
from yacs.config import CfgNode as CN
import yaml

class ConfigManager:
    """
    A class to manage configurations loaded from a YAML file.

    Explanation:
        This class provides methods to access, convert, save, and load configuration
        data from a YAML file.

    Args:
        None

    Returns:
        None
    """

    def __init__(self, config_file_path):
        """
        Constructor for ConfigManager.Initializes the ConfigManager 
        with the path to the YAML configuration file.

        Args:
            config_file_path (str): The path to the YAML configuration file.

        Returns:
            None
        """
        self.config_file_path = config_file_path
        self.cfg = self._load_cfg_from_yaml()

    def get_configs(self):
        """
        Get the configuration as a dictionary.

        Explanation:
            Returns a copy of the configuration as a dictionary, useful for accessing
            individual configuration values.

        Args:
            None

        Returns:
            dict: A dictionary representation of the configuration.
        """
        return self.cfg.clone()

    def convert_cfg_to_dict(self, cfg_node, key_list=[]):
        """
        Convert a config node to a dictionary recursively.

        Explanation:
            This method recursively converts a CfgNode instance (the internal
            representation of the config) to a regular Python dictionary.

        Args:
            cfg_node (CfgNode): The CfgNode instance to convert.
            key_list (list[str], optional): A list to track the current key path during
                recursive conversion. Defaults to an empty list.

        Returns:
            dict: The converted dictionary representation of the CfgNode.
        """
        VALID_TYPES = {tuple, list, str, int, float, bool}
        if not isinstance(cfg_node, CN):
            if type(cfg_node) not in VALID_TYPES:
                print(
                    f"Key {'.'.join(key_list)} with value {cfg_node} is not a valid type; valid types: {VALID_TYPES}"
                )
            return cfg_node
        else:
            cfg_dict = dict(cfg_node)
            for k, v in cfg_dict.items():
                cfg_dict[k] = self.convert_cfg_to_dict(v, key_list + [k])
            return cfg_dict

    def save_cfg_as_yaml(self, config_dict):
        """
        Save the configuration dictionary as a YAML file.

        Explanation:
            Saves the provided dictionary representation of the configuration to a YAML file
            at the specified path.

        Args:
            config_dict (dict): The dictionary representing the configuration to save.

        Returns:
            None
        """
        with open(f"{self.config_file_path}/configs.yaml", "w+") as yaml_file:
            yaml.safe_dump(config_dict, yaml_file, default_flow_style=False)

    def _load_cfg_from_yaml(self):
        """
        Load the configuration from a YAML file.

        Explanation:
            This method internally loads the configuration from the specified YAML file path
            and converts it to a CfgNode instance.

        Args:
            None

        Returns:
            CfgNode: The CfgNode instance representing the loaded configuration.
        """
        with open(f"{self.config_file_path}/configs.yaml", "r") as yaml_file:
            loaded = yaml.safe_load(yaml_file)
        return CN(loaded)



import random
import torch
import torch.nn as nn
from torchvision.transforms.functional import rotate
from torchvision.transforms import InterpolationMode


def inference_time_augmentation(data,i):
    """
    Apply inference-time augmentation to input data.

    Args:
        data (torch.Tensor): Input data with shape (channels, height, width).
        i (int): Index specifying the augmentation type:
            0: No flip
            1: Flip along the last dimension
            2: Flip along the second-to-last dimension
            3: Flip along both dimensions

    Returns:
        torch.Tensor: Augmented data.
    """
    if i == 0:
        x = data
    elif i == 1:
        x = torch.flip(data.clone(),dims=(-1,))
    elif i == 2:
        x = torch.flip(data.clone(),dims=(-2,))
    elif i == 3:
        x = torch.flip(data.clone(),dims=(-2,-1))
    
    return x


class TorchRandomRotate(nn.Module):
    """
    Apply random rotation to input images and masks. In contrast to the implementation in the Kornia library, this will apply a rotation by
    different degrees to each element instead of rotating all images in batch by the same degree.

    Args:
        degrees (tuple or list): Range of possible rotation angles.
        probability (float, default=1): Probability of applying the rotation.
        interpolation (InterpolationMode, default=BILINEAR): Interpolation mode for rotation.
        center (tuple, Optional): Center of rotation. If None, the center is the center of the image.
        fill (float, default=0): Value to fill the image during rotation.
        mask_fill (float, default=0): Value to fill the mask during rotation.

    Returns:
        tuple: Tuple containing the rotated image and mask (if provided).
        
    """
    
    def __init__(self, degrees, probability=1.0,interpolation=InterpolationMode.BILINEAR, center=None, fill=0,mask_fill=0):
        super().__init__()
        if not isinstance(degrees,(list,tuple)):
            degrees = (-abs(degrees),abs(degrees))

        self.degrees = degrees
        self.interpolation = interpolation
        self.center = center
        self.fill_value = fill
        self.mask_fill_value = mask_fill
        self.proba = probability

    @staticmethod
    def get_params(degrees) -> float:
        """
        Get parameters for a random rotation. Given a range of degrees "0 to 30 deg", it will generate a random degree in this range

        Parameters:
            degrees (tuple): A tuple representing the range of degrees for the rotation.

        Returns:
            float: A random angle within the specified degree range.

        """
        angle = float(torch.empty(1).uniform_(float(degrees[0]), float(degrees[1])).item())
        return angle
    

    @torch.no_grad()
    def __call__(self,img,mask=None):
        """
        Apply random rotation to the input image and mask. for each element in batch that isn't skipped, we generate an angle in a range,
        rotate the image by this angle, and apply rotation by the same angle to the mask (if it's not None),
        at the end we return the rotated image with the rotated mask (if provided)

        Parameters:
            img (torch.Tensor): Input image tensor of shape (batch_size, height, width, channels).
            mask (torch.Tensor, optional): Input mask tensor with same shape as image. Default is None.

        Returns:
            tuple: A tuple containing the rotated image tensor and the rotated mask tensor (if provided).

        Note:
            >>> If the mask is provided, both the image and mask are rotated by the same angle.
        """
        batch_size = img.shape[0]

        for i in range(batch_size):
            
            if random.random() > self.proba:
                continue

            angle = self.get_params(self.degrees)
            img[i,...] = rotate(img[i,...], angle, self.interpolation, False, self.center, self.fill_value)

            if mask is not None:
                mask[i,...] =  rotate(mask[i,...], angle, self.interpolation, False, self.center, self.mask_fill_value)
                mask = mask.float()
        if mask is not None:
            mask[mask<0] = self.mask_fill_value
        return img,mask


class RandomMaskIgnore(nn.Module):
    """
    RandomMaskIgnore class generates random masks within specified lengths with a given probability.
    It modifies the input mask tensor in-place, and the modified tensor is returned.

    Args:
        min_length (int, default=50): Minimum length of the randomly generated bounding box.
        max_length (int, default=100): Maximum length of the randomly generated bounding box.
        proba (float, default=0.5): Probability of applying the random mask modification.
        ignore_index (int, default=-10): Value used to fill the masked region.
    """

    def __init__(self,min_length=50,max_length=100,proba=0.5,ignore_index=-10):
        super().__init__()

        self.min_length = min_length
        self.max_length = max_length
        self.proba = proba
        self.ignore_index = ignore_index
    

    def generate_random_bbox(self,shape):
        """
        Generate a random bounding box within the specified shape.

        Args:
            shape (tuple): Height and width of the bounding box.

        Returns:
            tuple: Coordinates (top, left, bottom, right) of the bounding box.
        """
        H,W = shape
        L = random.randint(self.min_length,self.max_length)

        t = random.randint(0,H-L)
        b = t + L

        l = random.randint(0,W-L)
        r = l + L

        return (t,l,b,r)
    
    def mask_channel(self,bbox,channel):
        """
        Modify the specified channel by applying a mask within the bounding box.

        Args:
            bbox (tuple): Coordinates (top, left, bottom, right) of the bounding box.
            channel (tensor): Input channel to be modified.

        Returns:
            tensor: Modified channel.
        """
        (t,l,b,r) = bbox
        channel[:,t:b,l:r] = self.ignore_index
        return channel
    
    @torch.no_grad()
    def __call__(self,mask):
        """
        Apply random mask modifications to the input mask tensor with shape (Batch Size, Number of Channels, Height, Width)

        Args:
            mask (tensor): Input mask tensor.

        Returns:
            tensor: Modified mask tensor.
        """
        B,C,H,W = mask.shape
        for i in range(B):
            if random.random() > self.proba:
                continue
            bbox = self.generate_random_bbox((H,W))
            mask[i,...] = self.mask_channel(bbox,mask[i,...])
        
        return mask


class MaskPixelDrop(nn.Module):
    """
    MaskPixelDrop randomly drops pixels in the input mask tensor based on specified probabilities for positive and negative drops.
    
    Args:
        neg_drop (int or tuple, default for int =50): Probability range for dropping negative pixels. Default is (0, 50).
        pos_drop (int or tuple, default for int =50): Probability range for dropping positive pixels. Default is (0, 50).
        ignore_index (int, default=-10): Value used to fill the dropped pixels. Default is -10.

    Returns:
        torch.Tensor: Modified mask tensor with dropped pixels.
    """
    def __init__(self,neg_drop=50,pos_drop=50,ignore_index=-10):
        super().__init__()

        if not isinstance(neg_drop,tuple):
            neg_drop = (0,neg_drop)
        if not isinstance(pos_drop,tuple):
            pos_drop = (0,pos_drop)
        
        self.neg_drop = neg_drop
        self.pos_drop = pos_drop

        self.ignore_index = ignore_index
    
    @staticmethod
    def get_drop_proba(_range):
        """
        Get a random drop probability within the specified range.
        
        Args:
            _range (tuple): A tuple representing the range specified for the proba.
        
        Returns:
            int: A randomly generated probability in the range specified
        """
        return random.randint(_range[0],_range[1]) / 100
    
    def random_pixel_drop(self,gt,mask,_range):
        """
        Randomly drop pixels in the input mask tensor. Cs, Rs, Ws are column indices, row indices and depth indices respectively
        of non-zero elemtns in mask tensor.

        Args:
            gt (torch.Tensor): Ground truth tensor.
            mask (torch.Tensor): Input mask tensor.
            _range (tuple): Range for dropping pixels.

        Returns:
            torch.Tensor: Modified ground truth tensor with dropped pixels.
        """
        Cs,Hs,Ws = mask.nonzero(as_tuple=True)
        proba = self.get_drop_proba(_range)
        max_num = Cs.shape[0]
        drop_count = min(max_num,int(proba * max_num))
        
        if drop_count == 0 or max_num == 0:
            return gt

        indexes = random.sample(range(0, max_num), drop_count)
        Cs,Hs,Ws = Cs[indexes].tolist(),Hs[indexes].tolist(),Ws[indexes].tolist()
        gt[Cs,Hs,Ws] = self.ignore_index
        return gt

    @torch.no_grad()
    def __call__(self,mask):
        """
        Apply random pixel drops to the input mask tensor.

        Args:
            mask (torch.Tensor): Input mask tensor.

        Returns:
            torch.Tensor: Modified mask tensor with dropped pixels.
        """
        B,C,H,W = mask.shape
        pos_mask = mask.gt(0)
        neg_mask = mask.eq(0)
        for i in range(B):
            mask[i] = self.random_pixel_drop(mask[i],pos_mask[i],self.pos_drop)
            mask[i] = self.random_pixel_drop(mask[i],neg_mask[i],self.neg_drop)
        return mask



import numpy as np
import cv2
from PIL import Image
import os
import json
from shapely.geometry import Polygon, MultiPolygon
from skimage.draw import polygon
from skimage.morphology import erosion, square, binary_erosion
from skimage.io import imread, imsave
from tqdm import tqdm

import rasterio as rs
from cv2 import fillPoly, copyMakeBorder
from math import ceil
from typing import List, Tuple


class Masker:
    def __init__(
        self,
        out_size: Tuple[int, int] = (1024, 1024),
        erosion_kernel: str = 'cross',
        iterator_verbose: bool = True,
    ):
        r"""Initialize Masker object.

        Parameters
        ----------
        out_size : Tuple[int, int]
            Output size of the mask. Default is (1024, 1024).
        erosion_kernel : str
            Type of erosion kernel ('square' or 'cross'). Default is 'cross'.
        iterator_verbose : bool
            Flag to enable verbose mode for iterators. Default is True.
        """        
        self.sz = out_size
        self.pd_sz = (1044, 1044)
        self.x_off = ceil((self.pd_sz[1] - self.sz[1]) / 2)
        self.y_off = ceil((self.pd_sz[0] - self.sz[0]) / 2)
        assert (
            self.x_off >= 0 and self.y_off >= 0
        ), f'out size {self.sz} should be less than padded size {self.pd_sz}'
        assert (
            erosion_kernel.lower() in ['square', 'cross']
        ), f'erosion kernel type: [ {erosion_kernel} ] is not valid'
        self.ek_type = erosion_kernel.lower()
        self.itr_vrbs = iterator_verbose
        self.ldir = 'labels_match_pix'

    
    def load_labels(self, json_path: str) -> dict:
        r"""Load labels from a JSON file.

        Parameters
        ----------
        json_path : str
            Path to the JSON file containing labels.

        Returns
        -------
        dict
            Loaded labels from the JSON file.
        """        
        jfile = open(json_path, 'r')
        f = json.load(jfile)
        jfile.close()
        return f

    
    def poly_size(self, w: int, h: int) -> Polygon:
        r"""Create a Shapely Polygon representing a rectangle.

        Parameters
        ----------
        w : int
            Width of the rectangle.
        h : int
            Height of the rectangle.

        Returns
        -------
        Polygon
            Shapely Polygon representing the rectangle.
        """        
        return Polygon([[0, 0], [w - 1, 0], [w - 1, h - 1], [0, h - 1], [0, 0]])

    
    def get_strc(self) -> np.ndarray:
        r"""Get the erosion kernel structure based on the specified type.

        Returns
        -------
        np.ndarray
            Erosion kernel structure.
        """        
        if self.ek_type == 'square':
            return square(3)
        else:
            return np.array([[0, 1, 0], [1, 1, 1], [0, 1, 0]], dtype=np.uint8)

    
    def load_raster_file(self, raster_path: str) -> rs.DatasetReader:
        r"""Load a raster file using rasterio.

        Parameters
        ----------
        raster_path : str
            Path to the raster file.

        Returns
        -------
        rs.DatasetReader
            Rasterio dataset reader.
        """        
        return rs.open(raster_path)

    
    def get_img(self, raster_file: rs.DatasetReader) -> np.ndarray:
        r"""Get the image from a rasterio dataset reader.

        Parameters
        ----------
        raster_file : rs.DatasetReader
            Rasterio dataset reader.

        Returns
        -------
        np.ndarray
            Image array.
        """        
        return raster_file.read().transpose(1, 2, 0).astype(np.uint8)

    
    def project_poly(
        self,
        poly: List[Tuple[float, float]],
        frs: rs.DatasetReader,
        size: Tuple[int, int],
        x_off: int = 0,
        y_off: int = 0,
    ) -> List[Tuple[int, int]]:
        r"""Project a polygon onto the raster image.

        Parameters
        ----------
        poly : List[Tuple[float, float]]
            List of coordinates representing the polygon.
        frs : rs.DatasetReader
            Rasterio dataset reader.
        size : Tuple[int, int]
            Size of the image.
        x_off : int, optional
            X offset, by default 0.
        y_off : int, optional
            Y offset, by default 0.

        Returns
        -------
        List[Tuple[int, int]]
            List of projected coordinates.
        """        
        k = []
        for tup in poly:
            _x, _y = frs.index(*tup)
            _x += x_off
            _y += y_off
            k.append([_x, _y])

        polk = Polygon(k)
        if not polk.is_valid:
            polk = polk.buffer(0.01)
        poll = self.poly_size(*size)
        intr = poll.intersection(polk)
        verbs = intr.geom_type == 'Polygon'
        return list(intr.exterior.coords) if verbs else []

    
    def crop(self, img: np.ndarray, y_off: int, x_off: int, h: int, w: int) -> np.ndarray:
        r"""Crop an image.

        Parameters
        ----------
        img : np.ndarray
            Input image.
        y_off : int
            Y offset.
        x_off : int
            X offset.
        h : int
            Height of the cropped image.
        w : int
            Width of the cropped image.

        Returns
        -------
        np.ndarray
            Cropped image.
        """        
        return img[y_off : y_off + h, x_off : x_off + w]

    
    def make_mask_with_borders(
        self, polys: List[Polygon], size: Tuple[int, int] = (1024, 1024)
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        r"""Generate mask from polygons.

        Parameters
        ----------
        polys : List of polygons
            List of polygons.
        size : Tuple[int, int], optional
            Size of the mask, by default (1024, 1024).

        Returns
        -------
        Tuple[np.ndarray, np.ndarray, np.ndarray]
            Tuple containing instances, buildings, and borders masks.
        """        
        builds, border = [np.zeros(size, dtype=np.uint8), np.zeros(size, dtype=np.uint8)]
        instances = np.zeros(size, dtype=np.int32)
        strc = self.get_strc()
        itr = enumerate(polys)
        if self.itr_vrbs:
            itr = tqdm(itr)
            itr.set_description('generating mask')
        for i, mulpol in itr:
            for j, pol in enumerate(mulpol):
                arr_pol = np.array(pol, dtype=np.int32)
                hs, ws = polygon(arr_pol[:, 0], arr_pol[:, 1], size)
                instances[hs, ws, ...] = np.int32(i + 1) if (j == 0) else 0
            instance = instances == np.int32(i + 1)
            try:
                k = np.where(instance > 0)
                _t = k[0].min() - 2
                _l = k[1].min() - 2
                _b = k[0].max() + 2
                _r = k[1].max() + 2

                crop_instance = instance[_t:_b, _l:_r]                
                bld = binary_erosion(crop_instance, footprint=strc)
                brdr = bld ^ crop_instance
                brdr1 = np.zeros_like(instance, dtype=brdr.dtype)
                brdr1[_t:_b, _l:_r] = brdr
                border[brdr1 == True] = np.uint8(255)

            except:
                bld = binary_erosion(instance, footprint=strc)
                brdr = bld ^ instance
                border[brdr == True] = np.uint8(255)

        builds[instances > 0] = np.uint8(255)        
        return instances, builds, border


    def mask(self, raster_path: str, json_path: str) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        r"""Generate mask from raster image and labels.

        Parameters
        ----------
        raster_path : str
            Path to the raster image.
        json_path : str
            Path to the JSON file containing labels.

        Returns
        -------
        Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]
            Tuple containing the image, instances, buildings, and borders masks.
        """        
        raster = self.load_raster_file(raster_path)
        img = self.get_img(raster)
        js = self.load_labels(json_path)
        labels = js['features']
        polys = []

        for label in labels:
            multipoly = label['geometry']['coordinates']
            proj_multipoly = []
            for poly in multipoly:                
                mm = self.project_poly(poly, raster, self.pd_sz, self.x_off, self.y_off)
                if len(mm) > 0:
                    proj_multipoly.append(mm)
            polys.append(proj_multipoly)

        ins, b, br = self.make_mask_with_borders(polys, size=self.pd_sz)
        kwargs = {'y_off': self.y_off, 'x_off': self.x_off, 'h': self.sz[0], 'w': self.sz[1]}
        ins = self.crop(ins, **kwargs)
        b = self.crop(b, **kwargs)
        br = self.crop(br, **kwargs)
        img = cv2.copyMakeBorder(
            img,
            top=0,
            bottom=self.sz[0] - img.shape[0],
            left=0,
            right=self.sz[1] - img.shape[1],
            borderType=cv2.BORDER_CONSTANT,
            value=0,
        )
        return img, ins, b, br

    
    def _collect(self, labels: dict) -> dict:
        r"""Collect metadata from labels.

        Parameters
        ----------
        labels : dict
            Dictionary containing labels.

        Returns
        -------
        dict
            Metadata collected from labels.
        """        
        _meta = {}
        for label in labels['features']:
            pid = str(np.int32(label['properties']['Id']))
            _meta[pid] = {}
            _meta[pid]['area'] = label['properties']['area']
            _meta[pid]['geometry'] = label['geometry']['coordinates']
        
        return _meta
    
    
    def int_coords(self, x: float) -> np.int32:
        r"""Convert coordinates to integer.

        Parameters
        ----------
        x : float
            Input coordinate.

        Returns
        -------
        np.int32
            Integer coordinate.
        """        
        return np.array(x).round().astype(np.int32)

    
    def instances(self, size: Tuple[int, int], labels: dict) -> np.ndarray:
        r"""Generate instances mask from labels.

        Parameters
        ----------
        size : Tuple[int, int]
            Size of the mask.
        labels : dict
            Dictionary containing labels.

        Returns
        -------
        np.ndarray
            Instances mask.
        """        
        ins_mask = np.zeros(size, dtype=np.int32)

        for pid, d in labels.items():
            int_id = np.int32(pid)
            polys = d['geometry']
            for i, poly in enumerate(polys):
                poly.append(poly[0])
                S = Polygon(poly)
                PS = self.poly_size(size[1], size[0])
                S = S.intersection(PS)
                Stype = S.geom_type

                if Stype == 'Polygon':
                    arr_pol = self.int_coords(S.exterior.coords)
                    if len(arr_pol.shape) != 2:
                        continue
                    ws, hs = polygon(arr_pol[:, 0], arr_pol[:, 1], size)
                    ins_mask[hs, ws, ...] = int_id if i == 0 else 0

                elif Stype == 'MultiPolygon':
                    for s in S:
                        arr_pol = self.int_coords(s.exterior.coords)
                        if len(arr_pol.shape) != 2:
                            continue
                        ws, hs = polygon(arr_pol[:, 0], arr_pol[:, 1], size)
                        ins_mask[hs, ws, ...] = int_id if i == 0 else 0

                else:
                    for point in list(S.coords):
                        x, y = list(map(np.int32, point))
                        ins_mask[y, x, ...] = int_id if i == 0 else 0

        return ins_mask

    
    # def borders(self, ins_mask: np.ndarray) -> np.ndarray:
    #     r"""Generate borders mask from instances mask.

    #     Parameters
    #     ----------
    #     ins_mask : np.ndarray
    #         Instances mask.

    #     Returns
    #     -------
    #     np.ndarray
    #         Borders mask.
    #     """        
    #     ins_borders = np.zeros_like(ins_mask,dtype = np.int32)
    #     ids = sorted(np.unique(ins_mask))[1:]
    #     strc = self.get_strc()
    #     for iid in ids:
    #         instance = ins_mask == iid
    #         try:
    #             k=np.where(instance>0)
    #             _t = k[0].min() - 3
    #             _l = k[1].min() - 3
    #             _b = k[0].max() + 3
    #             _r = k[1].max() + 3
                
    #             crop_instance = instance[_t:_b,_l:_r]
    #             bld = binary_erosion(crop_instance, strc)
    #             brdr = bld ^ crop_instance
    #             brdr1 = np.zeros_like(instance,dtype=brdr.dtype)
    #             brdr1[_t:_b,_l:_r] =brdr
    #             ins_borders[brdr1 == True] = iid
                
    #         except:
    #             bld = binary_erosion(instance, strc)
    #             brdr = bld ^ instance
    #             ins_borders[brdr == True] = iid
    #     return ins_borders

    
    def to_rgb(self, img: np.ndarray) -> np.ndarray:
        r"""Convert an image to RGB format.

        Parameters
        ----------
        img : np.ndarray
            Input image array with shape (H, W, C).

        Returns
        -------
        np.ndarray
            RGB image array with shape (H, W, 3).
        """        
        rgb = np.ascontiguousarray(img[...,:3],dtype = np.uint8)
        return rgb

    
    def to_gray(self, mask: np.ndarray) -> np.ndarray:
        r"""Convert a mask to grayscale.

        Parameters
        ----------
        mask : np.ndarray
            Input mask array with shape (H, W).

        Returns
        -------
        np.ndarray
            Grayscale mask array with values 0 or 255.
        """        
        return (mask>0).astype(np.uint8) * 255
    
    
    def generate_dataset(self, save_path: str, data: str) -> None:
        r"""Generate a dataset with images, building masks, and border masks.

        Parameters
        ----------
        save_path : str
            Path to save the generated dataset.
        data : str
            Path to the input data directory.

        Returns
        -------
        None
        """        
        os.mkdir(save_path)
        ids = sorted(os.listdir(data))
        for iid in ids[42:]:
            imgs_save = f'{save_path}/{iid}'
            imgs_path = f'{self.data}/{iid}/images_masked'
            labels_path = f'{self.data}/{iid}/{self.ldir}'
            lod = os.listdir(imgs_path)
            loader = tqdm(lod) if(self.itr_vrbs) else lod
            loader.set_description(f'{iid}')
            os.mkdir(imgs_save)
            for exten in loader:
                e = exten.split('.tif')[0]
                img_save = f'{imgs_save}/{e}'
                os.mkdir(img_save)
                
                img_rgba = imread(f'{imgs_path}/{e}.tif')
                img_rgb = self.to_rgb(img_rgba)
                
                lf = self.load_labels(iid, e)
                labels = self._collect(lf)
                
                size = img_rgb.shape[:2]
                ins_mask = self.instances(size, labels)
                ins_borders = self.borders(ins_mask)
                
                imsave(f'{img_save}/image.png',img_rgb)
                imsave(f'{img_save}/buildings.png',self.to_gray(ins_mask))
                imsave(f'{img_save}/borders.png',self.to_gray(ins_borders))
                np.save(f'{img_save}/buildings.npy',ins_mask)
                np.save(f'{img_save}/borders.npy',ins_borders)



from .radam import PlainRAdam,RAdam
from .novograd import Novograd
from .ranger import Ranger
from .ralamb import Ralamb
from .rangerlars import RangerLars
from .lookahead import Lookahead,LookaheadAdam
#from .lamb import Lamb
from .diffgrad import DiffGrad
from .adamod import AdaMod
from .madam import Madam
from .apollo import Apollo
from .adabelief import AdaBelief
from .madgrad import MADGRAD
from .adan import Adan


from torch.optim.lr_scheduler import LambdaLR
# polylr

class PolyLR_WWP(LambdaLR):
    def __init__(self, optimizer, epochs, warmup,ratio=0.9):
        warmup = min(epochs,max(0,warmup-1))
        decay_epochs = epochs - warmup
        xlambda = lambda x : 1.0 if(x<warmup) else (1 - ((x - warmup) / decay_epochs ) ** ratio)
        super().__init__(optimizer, xlambda)

class PolyLR(PolyLR_WWP):
    def __init__(self, optimizer, epochs, ratio=0.9):
        super().__init__(optimizer, epochs,0,ratio)


import numpy as np
import random
from PIL import Image
import Dataset
import torch

from albumentations import Resize,HorizontalFlip,VerticalFlip,Compose,OneOf,NoOp,\
                            Cutout,InvertImg,Normalize,ChannelShuffle,\
                            HueSaturationValue,GaussianBlur,GaussNoise,MultiplicativeNoise,ShiftScaleRotate,Normalize,CLAHE,ColorJitter,RandomBrightnessContrast

pos_augs = OneOf([
                  VerticalFlip(p=1.0),
                   HorizontalFlip(p=1.0),
                   Compose([
                            VerticalFlip(1.0),
                            HorizontalFlip(p=1.0),
                                
                     ],p=1.0),
                ],p=0.8)

color_augs = OneOf([
                    #ChannelShuffle(p=1.0),
                    HueSaturationValue(hue_shift_limit=15, sat_shift_limit=15, val_shift_limit=15,p=1.0),
                    CLAHE(p=1.0),
                    ColorJitter(brightness=0.2, contrast=0.2, saturation=0.2, hue=0.2,p=1.0) ,
                    RandomBrightnessContrast(brightness_limit=0.2, contrast_limit=0.2, brightness_by_max=True, p=1.0)
                ],p=0.4)

hard_augs = OneOf([
                   GaussianBlur(blur_limit=3,p=1.0),
                   GaussNoise(var_limit=(100.0, 200.0),mean=0,p=1.0),
                   Cutout(num_holes=4, max_h_size=30, max_w_size=35, fill_value=[0],p=1.0),
                   MultiplicativeNoise(multiplier=(0.7, 1.1), per_channel=True,p=1.0)
              ],p=0.3)

norm_aug = Normalize(mean=(0.485, 0.456, 0.406), 
                     std=(0.229, 0.224, 0.225), 
                     max_pixel_value=255.0, 
                     always_apply=True,
                     p=1.0)
tfm0 = NoOp()


tfm1 = Compose([
                norm_aug
                ],p=1.0)

tfm2 = Compose([
                pos_augs,
                norm_aug
              ],p=1.0)

tfm3 = Compose([
                pos_augs,
                color_augs,
                norm_aug
              ],p=1.0)

tfm4 = Compose([
                pos_augs,
                color_augs,
                hard_augs,
                norm_aug
              ],p=1.0)

def get_tfm(number):
  if(number == 0):
    return tfm0
  elif(number == 1):
    return tfm1
  elif(number == 2):
    return tfm2
  elif(number == 3):
    return tfm3
  elif(number == 4):
    return tfm4
  else:
    raise ValueError('Transform {} is not an option!!'.format(number))
  
  

class FloodNetDataset_FS(Dataset):
  def __init__(self,data_path,classes = 11,ignore_class = 10,fold=1,train=True,transform_number=2,to_tensor=False,shuffle=False):
    self.data = []
    self.train = train
    self.fold = fold
    self.tfm = get_tfm(transform_number)
    self.tt = to_tensor
    self.shuff = shuffle
    self.classes = classes
    self.ig_class = ignore_class
    self.class_names = ['background' , 'building_flooded' ,'building_non-flooded' ,
                        'road_flooded' , 'road_non-flooded' , 'water' , 'tree' ,
                        'vehicle' , 'pool' ,  'grass']
    df = pd.read_csv(f'{data_path}/folds.csv')

    if(train):
      df = df.loc[(df['fold'] != self.fold)]
    else: 
      df = df.loc[(df['fold'] == self.fold)]

    for i,row in df.iterrows():
      iid = row['iid']
      #if(iid == 6707):print(len(self.data))
      clss = row['class']
      self.data.append({
          'iid' : row['iid'],
          'class' : 0.0 if(clss == 'non-flooded') else 1.0,
          'img_path' : f'{data_path}/labeled/images/{iid}_{clss}.png',
          'mask_path' : f'{data_path}/labeled/masks/{iid}_{clss}_mask.png',
      })

    if(self.shuff):
      random.shuffle(self.data)
    self.regen()

  def __getitem__(self,index):
    self.idx = index
    curr = self.__get_current_data__()
    img = np.array(Image.open(curr['img_path']),dtype=np.uint8)
    mask = np.array(Image.open(curr['mask_path']),dtype=np.uint8)

    aug = self.tfm(image=img,mask=mask)
    img,mask = aug['image'],aug['mask']

    masks = []
    h,w = mask.shape
    ig_mask = None

    for i in range(self.classes):
      m = np.zeros((h,w),dtype=np.uint8)
      if(i == self.ig_class):
        m[mask != i] = 1
        ig_mask = m
      else:
        m[mask == i] = 1
        masks.append(m)

    mask = np.stack(masks,axis = -1).astype(np.uint8)
    label = np.ones(1) * curr['class']
    if(self.tt):
      img = torch.from_numpy(img.transpose(2,0,1)).float()
      mask = torch.from_numpy(mask.transpose(2,0,1)).float()
      ig_mask = torch.from_numpy(ig_mask).unsqueeze(0).float()
      label = torch.from_numpy(label).float()

    return img,mask,ig_mask,label

  def regen(self):
    self.idx = 0

  def change_tfm(self,tfm):
    self.tfm = tfm

  def change_tt(self,tt):
    self.tt = tt
    
  def __get_current_data__(self):
    return self.data[self.idx]
    
  def __len__(self):
    return len(self.data)
     


import rasterio
import os
from rasterio.windows import Window
import geopandas as gpd
from shapely.geometry import Polygon,box
from PIL import Image


class CropGeoTiff:

    """
    The CropGeoTiff class is designed to efficiently crop large TIFF images 
    along with their corresponding shapefiles into smaller, uniform-sized images.
    This class provides two essential functions: `crop_images` for cropping georeferenced 
    images and `crop_shapefiles` for cropping georeferenced shapefiles. Additionally, 
    the `crop_nongeoref_images_shapefile` function is specifically crafted to handle the cropping 
    of non-georeferenced images and shapefiles.
    """
    def __init__(self,fragment_size=(1024, 1024)):
        """
         @param fragment_size = fragment size cropped
        """

        self.fragment_size = fragment_size 


    def crop_images(self,input_dir,output_dir):
        """
        @param input_dir = path of the directory containing images to be cropped
        @param output_dir = path of directory for saving cropped images
        """

        for i in os.listdir(input_dir):
            name = i.split('.')[0]
            input_path = os.path.join(input_dir, i)

            with rasterio.open(input_path) as src:
                original_profile = src.profile

                num_rows = src.height // self.fragment_size[0]
                num_cols = src.width // self.fragment_size[1]
                ind = 0

                for row in range(num_rows):
                    for col in range(num_cols):
                   
                        window = Window(col * self.fragment_size[1], row * self.fragment_size[0], self.fragment_size[1], self.fragment_size[0])

                      
                        fragment = src.read(window=window)

                
                        fragment_profile = original_profile.copy()
                        fragment_profile['width'] = self.fragment_size[1]
                        fragment_profile['height'] = self.fragment_size[0]

                        
                        fragment_profile['transform'] = rasterio.windows.transform(window, src.transform)

            
                        save_subdir = os.path.join(output_dir, name)
                        os.makedirs(save_subdir, exist_ok=True)

                      
                        fragment_output_path = os.path.join(save_subdir, f'{name}_{ind}.tif')
                        with rasterio.open(fragment_output_path, 'w', **fragment_profile) as dst:
                            dst.write(fragment)
                        ind += 1




    def crop_shapefiles(self,images,shapefile_path,output_dir):
        """
        @param images = path of the diectory containing cropped images
        @param shapefile_path = path of the directory containing shapefiles to be cropped
        @param output_dir = path of the directory for saving cropped shapefiles
        """
        gdf = gpd.read_file(shapefile_path)
        for dirr in os.listdir(images): 
            for i in os.listdir(os.path.join(images,dirr)):
                name=dirr
                n=i.split('.')[0]
                with rasterio.open(images+'/'+name+'/'+i) as src:
                    image_bounds = src.bounds
                
                
                
                bbox = (image_bounds.left, image_bounds.bottom, image_bounds.right, image_bounds.top)
                cropped_gdf = gpd.clip(gdf, mask=bbox)
            
               
                output = output_dir+f"/{name}/{n}/"   
                os.makedirs(output,exist_ok=True)
                output_filename = f"{n}.shp"
                output_path = output + output_filename
                cropped_gdf.to_file(output_path)
                





    def crop_nongeoref_images_shapefile(self,input_dir,shapefile_path,output_image_dir,output_shp_dir):
        """
        @param input_dir = path of the directory containing images to be cropped
        @param shapefile_path = path of the directory containing shapefiles to be cropped
        @param output_image_dir = path of the directory to save cropped images
        @param output_shp_dir = path of the directory to save cropped shapefiles
        """
       
        gdf = gpd.read_file(shapefile_path)
        for i in os.listdir(input_dir):
            name = i.split('.')[0]
            input_path = os.path.join(input_dir, i)

            image = Image.open(input_path)
            width, height = image.size

            num_rows = height // self.fragment_size[1]
            num_cols = width // self.fragment_size[0]
            ind = 0

            for row in range(num_rows):
                for col in range(num_cols):
                    left = col * self.fragment_size[0]
                    upper = row * self.fragment_size[1]
                    right = left + self.fragment_size[0]
                    lower = upper + self.fragment_size[1]

                    fragment = image.crop((left, upper, right, lower))
                    bbox = box(left, upper, right, lower)

                    cropped_gdf = gdf[gdf.intersects(bbox)]
                    
                    save_subdir = os.path.join(output_image_dir, name)
                    os.makedirs(save_subdir, exist_ok=True)

                    fragment_output_path = os.path.join(save_subdir, f'{name}_{ind}.png')
                    fragment.save(fragment_output_path)
                    output_shp = output_shp_dir+f"/{name}/{name}_{ind}/"
                    os.makedirs(output_shp, exist_ok=True)
                    output_filename = f"{name}_{ind}.shp"
                    output_path = os.path.join(output_shp, output_filename)
                    cropped_gdf.to_file(output_path)
                
                    ind += 1



import torch
import math
from torch.optim import Optimizer
#AdamW
class AdamW(Optimizer):
    """Implements AdamW algorithm.
    It has been proposed in `Fixing Weight Decay Regularization in Adam`_.
    Arguments:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float, optional): learning rate (default: 1e-3)
        betas (Tuple[float, float], optional): coefficients used for computing
            running averages of gradient and its square (default: (0.9, 0.999))
        eps (float, optional): term added to the denominator to improve
            numerical stability (default: 1e-8)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
    .. Fixing Weight Decay Regularization in Adam:
    https://arxiv.org/abs/1711.05101
    """

    def __init__(self, params, lr=1e-3, betas=(0.9, 0.999), eps=1e-8,
                 weight_decay=0):
        defaults = dict(lr=lr, betas=betas, eps=eps,
                        weight_decay=weight_decay)
        super(AdamW, self).__init__(params, defaults)

    def step(self, closure=None):
        """Performs a single optimization step.
        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError('AdamW does not support sparse gradients, please consider SparseAdam instead')

                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state['step'] = 0
                    # Exponential moving average of gradient values
                    state['exp_avg'] = torch.zeros_like(p.data)
                    # Exponential moving average of squared gradient values
                    state['exp_avg_sq'] = torch.zeros_like(p.data)

                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                beta1, beta2 = group['betas']

                state['step'] += 1

                # according to the paper, this penalty should come after the bias correction
                # if group['weight_decay'] != 0:
                #     grad = grad.add(group['weight_decay'], p.data)

                # Decay the first and second moment running average coefficient
                exp_avg.mul_(beta1).add_(1 - beta1, grad)
                exp_avg_sq.mul_(beta2).addcmul_(1 - beta2, grad, grad)

                denom = exp_avg_sq.sqrt().add_(group['eps'])

                bias_correction1 = 1 - beta1 ** state['step']
                bias_correction2 = 1 - beta2 ** state['step']
                step_size = group['lr'] * math.sqrt(bias_correction2) / bias_correction1

                # w = w - wd * lr * w
                if group['weight_decay'] != 0:
                    p.data.add_(-group['weight_decay'] * group['lr'], p.data)

                # w = w - lr * w.grad
                p.data.addcdiv_(-step_size, exp_avg, denom)

                # w = w - wd * lr * w - lr * w.grad
                # See http://www.fast.ai/2018/07/02/adam-weight-decay/

        return loss


# original repo https://github.com/lancopku/AdaMod/blob/master/adamod/adamod.py
import math
import torch
from torch.optim import Optimizer

class AdaMod(Optimizer):
    """Implements AdaMod algorithm with Decoupled Weight Decay (arxiv.org/abs/1711.05101)
    It has been proposed in `Adaptive and Momental Bounds for Adaptive Learning Rate Methods`_.
    Arguments:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float, optional): learning rate (default: 1e-3)
        betas (Tuple[float, float], optional): coefficients used for computing
            running averages of gradient and its square (default: (0.9, 0.999))
        beta3 (float, optional): smoothing coefficient for adaptive learning rates (default: 0.9999)
        eps (float, optional): term added to the denominator to improve
            numerical stability (default: 1e-8)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
    """

    def __init__(self, params, lr=1e-3, betas=(0.9, 0.999), beta3=0.999,
                 eps=1e-8, weight_decay=0):
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))
        if not 0.0 <= beta3 < 1.0:
            raise ValueError("Invalid beta3 parameter: {}".format(beta3))
        defaults = dict(lr=lr, betas=betas, beta3=beta3, eps=eps,
                        weight_decay=weight_decay)
        super(AdaMod, self).__init__(params, defaults)

    def __setstate__(self, state):
        super(AdaMod, self).__setstate__(state)

    def step(self, closure=None):
        """Performs a single optimization step.
        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError(
                        'AdaMod does not support sparse gradients')

                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state['step'] = 0
                    # Exponential moving average of gradient values
                    state['exp_avg'] = torch.zeros_like(p.data)
                    # Exponential moving average of squared gradient values
                    state['exp_avg_sq'] = torch.zeros_like(p.data)
                    # Exponential moving average of actual learning rates
                    state['exp_avg_lr'] = torch.zeros_like(p.data)

                exp_avg, exp_avg_sq, exp_avg_lr = state['exp_avg'], state['exp_avg_sq'], state['exp_avg_lr']
                beta1, beta2 = group['betas']

                state['step'] += 1

                # Decay the first and second moment running average coefficient
                exp_avg.mul_(beta1).add_(1 - beta1, grad)
                exp_avg_sq.mul_(beta2).addcmul_(1 - beta2, grad, grad)

                denom = exp_avg_sq.sqrt().add_(group['eps'])

                bias_correction1 = 1 - beta1 ** state['step']
                bias_correction2 = 1 - beta2 ** state['step']
                step_size = group['lr'] * math.sqrt(bias_correction2) / bias_correction1

                if group['weight_decay'] != 0:
                    p.data.add_(-group['weight_decay'] * group['lr'], p.data)

                # Applies momental bounds on actual learning rates
                step_size = torch.full_like(denom, step_size)
                step_size.div_(denom)
                exp_avg_lr.mul_(group['beta3']).add_(1 - group['beta3'], step_size)
                step_size = torch.min(step_size,  exp_avg_lr)
                step_size.mul_(exp_avg)

                p.data.add_(-step_size)

        return loss



import torch
from icecream import ic
import geopandas as gpd
from tmp import convert_polygon_to_mask_batch,binary_mask_to_polygon
from segmentation_models_pytorch.utils.metrics import Accuracy,Recall
from segmentation_models_pytorch.utils.base import Metric


def _threshold(x, threshold=0.5):
    """
    Thresholds the input tensor based on a specified threshold value.

    Args:
        x (torch.Tensor): Input tensor to be thresholded.
        threshold (float, optional): The threshold value. If None, no threshold is applied.
                                     Default is 0.5.

    Returns:
        torch.Tensor: Thresholded tensor where values greater than or equal to the threshold
                      are set to 1, and values below the threshold remain unchanged.

    Note:
        - The input tensor `x` and the returned tensor will have the same dtype.
        - If `threshold` is None, the input tensor `x` is returned unchanged.
    """
    if threshold is not None:
        return (x >= threshold).type(x.dtype)
    else:
        return x

class DiceScore:
    """
    Initializes an instance of the DiceScore class.

    Args:
        threshold (float, optional): Threshold value for thresholding predictions. Default is 0.5.
        ignore_index (int, optional): Index to ignore in ground truth. Default is -100.
    """
    def __init__(self,threshold = 0.5,ignore_index=-100):
        self.thresh = threshold
        self.eps = 1e-6
        self.dice_score_sum = 0.0
        self.weights_sum = 0.0
        self.ignore_index = ignore_index

    @torch.no_grad()
    def __call__(self, y_pr,y_gt,thresh_gt=False):
        """
        Calculates the Dice Score between predicted and ground truth tensors.

        Args:
            y_pr (torch.Tensor): Predicted tensor.
            y_gt (torch.Tensor): Ground truth tensor.
            thresh_gt (bool, optional): Whether to threshold the ground truth tensor. Default is False.

        Returns:
            torch.Tensor: Dice score between the predicted and ground truth tensors.

        Note:
            - This function accumulates dice scores and weights over multiple calls.
            - The final dice score is computed as the sum of dice scores divided by the sum of weights.
        """
        b,c,_,_ = y_pr.shape
        y_pr = _threshold(y_pr,self.thresh)
        if thresh_gt:
            y_gt = _threshold(y_gt,self.thresh)
        
        if self.ignore_index is not None:
            y_nindex = (y_gt != self.ignore_index).type(y_gt.dtype)
            y_gt = y_gt * y_nindex
            y_pr = y_pr * y_nindex

        y_gt = y_gt.view(b,c,-1).float() # >> B,C,HxW
        y_pr = y_pr.view(b,c,-1).float() # >> B,C,HxW


        intersection = torch.sum(y_pr * y_gt, dim=-1)
        cardinality = torch.sum(y_pr + y_gt, dim=-1)

        dice_score = (2.0 * intersection)/ cardinality.clamp_min(self.eps)
        
        weights = 1.0 - (cardinality == 0.0).float()
        weights = weights.sum(dim=-1)

        dice_score = dice_score.sum(dim=-1) / weights.clamp_min(1.0)


        self.dice_score_sum += dice_score.sum()
        self.weights_sum += (weights > 0.0).float().sum()

        dice_score = self.dice_score_sum / self.weights_sum.clamp_min(1.0)


        return dice_score

class IoUScore:
    """
    Initializes an instance of the IoUScore class.

    Args:
        threshold (float, optional): Threshold value for thresholding predictions. Default is 0.5.
        ignore_index (int, optional): Index to ignore in ground truth. Default is -100.
    """
    def __init__(self,threshold = 0.5,ignore_index=-100):
        self.thresh = threshold
        self.eps = 1e-6
        self.iou_score_sum = 0.0
        self.weights_sum = 0.0
        self.ignore_index = ignore_index

    @torch.no_grad()
    def __call__(self, y_pr,y_gt,thresh_gt=False):
        """
        Calculates the Intersection over Union (IoU) score between predicted and ground truth tensors.

        Args:
            y_pr (torch.Tensor): Predicted tensor.
            y_gt (torch.Tensor): Ground truth tensor.
            thresh_gt (bool, optional): Whether to threshold the ground truth tensor. Default is False.

        Returns:
            torch.Tensor: IoU score between the predicted and ground truth tensors.

        Note:
            - This function accumulates IoU scores and weights over multiple calls.
            - The final IoU score is computed as the sum of IoU scores divided by the sum of weights.
        """
        b,c,_,_ = y_pr.shape
        y_pr = _threshold(y_pr,self.thresh)
        if thresh_gt:
            y_gt = _threshold(y_gt,self.thresh)
        
        if self.ignore_index is not None:
            y_nindex = (y_gt != self.ignore_index).type(y_gt.dtype)
            y_gt = y_gt * y_nindex
            y_pr = y_pr * y_nindex

        y_gt = y_gt.view(b,c,-1).float() # >> B,C,HxW
        y_pr = y_pr.view(b,c,-1).float() # >> B,C,HxW


        intersection = torch.sum(y_pr * y_gt, dim=-1)
        union = torch.sum((y_pr + y_gt).clamp_max(1.0), dim=-1)

        iou_score =  intersection / union.clamp_min(self.eps)
        
        weights = 1.0 - (union == 0.0).float()
        weights = weights.sum(dim=-1)

        iou_score = iou_score.sum(dim=-1) / weights.clamp_min(1.0)


        self.iou_score_sum += iou_score.sum()
        self.weights_sum += (weights > 0.0).float().sum()

        iou_score = self.iou_score_sum / self.weights_sum.clamp_min(1.0)

        return iou_score



# from https://raw.githubusercontent.com/facebookresearch/madgrad/master/madgrad/madgrad.py
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import math
from typing import TYPE_CHECKING, Any, Callable, Optional

import torch
import torch.optim

if TYPE_CHECKING:
    from torch.optim.optimizer import _params_t
else:
    _params_t = Any


class MADGRAD(torch.optim.Optimizer):
    """
    MADGRAD_: A Momentumized, Adaptive, Dual Averaged Gradient Method for Stochastic 
    Optimization.

    .. _MADGRAD: https://arxiv.org/abs/2101.11075

    MADGRAD is a general purpose optimizer that can be used in place of SGD or
    Adam may converge faster and generalize better. Currently GPU-only.
    Typically, the same learning rate schedule that is used for SGD or Adam may
    be used. The overall learning rate is not comparable to either method and
    should be determined by a hyper-parameter sweep.

    MADGRAD requires less weight decay than other methods, often as little as
    zero. Momentum values used for SGD or Adam's beta1 should work here also.

    On sparse problems both weight_decay and momentum should be set to 0.

    Arguments:
        params (iterable): 
            Iterable of parameters to optimize or dicts defining parameter groups.
        lr (float): 
            Learning rate (default: 1e-2).
        momentum (float): 
            Momentum value in  the range [0,1) (default: 0.9).
        weight_decay (float): 
            Weight decay, i.e. a L2 penalty (default: 0).
        eps (float): 
            Term added to the denominator outside of the root operation to improve numerical stability. (default: 1e-6).
    """

    def __init__(
        self, params: _params_t, lr: float = 1e-2, momentum: float = 0.9, weight_decay: float = 0, eps: float = 1e-6,
    ):
        if momentum < 0 or momentum >= 1:
            raise ValueError(f"Momentum {momentum} must be in the range [0,1]")
        if lr <= 0:
            raise ValueError(f"Learning rate {lr} must be positive")
        if weight_decay < 0:
            raise ValueError(f"Weight decay {weight_decay} must be non-negative")
        if eps < 0:
            raise ValueError(f"Eps must be non-negative")

        defaults = dict(lr=lr, eps=eps, momentum=momentum, weight_decay=weight_decay, k=0)
        super().__init__(params, defaults)

        for group in self.param_groups:
            for p in group["params"]:
                state = self.state[p]

                state["grad_sum_sq"] = torch.zeros_like(p.data).detach()
                state["s"] = torch.zeros_like(p.data).detach()
                if momentum != 0:
                    state["x0"] = torch.clone(p.data).detach()

    @property
    def supports_memory_efficient_fp16(self) -> bool:
        return False

    @property
    def supports_flat_params(self) -> bool:
        return True

    def step(self, closure: Optional[Callable[[], float]] = None) -> Optional[float]:
        """Performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            eps = group["eps"]
            k = group["k"]
            lr = group["lr"] + eps
            decay = group["weight_decay"]
            momentum = group["momentum"]

            ck = 1 - momentum
            lamb = lr * math.pow(k + 1, 0.5)

            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad.data
                state = self.state[p]

                if momentum != 0.0 and grad.is_sparse:
                    raise RuntimeError("momentum != 0 is not compatible with sparse gradients")

                grad_sum_sq = state["grad_sum_sq"]
                s = state["s"]

                # Apply weight decay
                if decay != 0:
                    if grad.is_sparse:
                        raise RuntimeError("weight_decay option is not compatible with sparse gradients")

                    grad.add_(p.data, alpha=decay)

                if grad.is_sparse:
                    grad = grad.coalesce()
                    grad_val = grad._values()

                    p_masked = p.sparse_mask(grad)
                    grad_sum_sq_masked = grad_sum_sq.sparse_mask(grad)
                    s_masked = s.sparse_mask(grad)

                    # Compute x_0 from other known quantities
                    rms_masked_vals = grad_sum_sq_masked._values().pow(1 / 3).add_(eps)
                    x0_masked_vals = p_masked._values().addcdiv(s_masked._values(), rms_masked_vals, value=1)

                    # Dense + sparse op
                    grad_sq = grad * grad
                    grad_sum_sq.add_(grad_sq, alpha=lamb)
                    grad_sum_sq_masked.add_(grad_sq, alpha=lamb)

                    rms_masked_vals = grad_sum_sq_masked._values().pow_(1 / 3).add_(eps)

                    s.add_(grad, alpha=lamb)
                    s_masked._values().add_(grad_val, alpha=lamb)

                    # update masked copy of p
                    p_kp1_masked_vals = x0_masked_vals.addcdiv(s_masked._values(), rms_masked_vals, value=-1)
                    # Copy updated masked p to dense p using an add operation
                    p_masked._values().add_(p_kp1_masked_vals, alpha=-1)
                    p.data.add_(p_masked, alpha=-1)
                else:
                    if momentum == 0:
                        # Compute x_0 from other known quantities
                        rms = grad_sum_sq.pow(1 / 3).add_(eps)
                        x0 = p.data.addcdiv(s, rms, value=1)
                    else:
                        x0 = state["x0"]

                    # Accumulate second moments
                    grad_sum_sq.addcmul_(grad, grad, value=lamb)
                    rms = grad_sum_sq.pow(1 / 3).add_(eps)

                    # Update s
                    s.data.add_(grad, alpha=lamb)

                    # Step
                    if momentum == 0:
                        p.data.copy_(x0.addcdiv(s, rms, value=-1))
                    else:
                        z = x0.addcdiv(s, rms, value=-1)

                        # p is a moving average of z
                        p.data.mul_(1 - ck).add_(z, alpha=ck)

            group["k"] = group["k"] + 1
        return loss



import torch, math
from torch.optim.optimizer import Optimizer
import itertools as it
from .lookahead import *
from .ralamb import * 

# RAdam + LARS + LookAHead

# Lookahead implementation from https://github.com/lonePatient/lookahead_pytorch/blob/master/optimizer.py
# RAdam + LARS implementation from https://gist.github.com/redknightlois/c4023d393eb8f92bb44b2ab582d7ec20

def RangerLars(params, alpha=0.5, k=6, *args, **kwargs):
     ralamb = Ralamb(params, *args, **kwargs)
     return Lookahead(ralamb, alpha, k)



#visualization utils
def mask2rgb(mask,max_value=1.0):
    shape = mask.shape
    if len(shape) == 2:
        mask = mask[:,:,np.newaxis]
    h,w,c = mask.shape
    if c == 3:
        return mask
    if c == 4:
        return mask[:,:,:3]
    
    if c > 4:
        raise ValueError
    
    padded = np.zeros((h,w,3),dtype=mask.dtype)
    padded[:,:,:c] = mask
    padded = (padded * max_value).astype(np.uint8)
    
    return padded


def make_rgb_mask(mask,color=(255,0,0)):
    h,w = mask.shape[:2]
    rgb = np.zeros((h,w,3),dtype=np.uint8)
    rgb[mask == 1.0,:] = color
    return rgb

def overlay_rgb_mask(img,mask,sel,alpha):

    sel = sel == 1.0
    img[sel,:] = img[sel,:] * (1.0 - alpha) + mask[sel,:] * alpha
    return img

def overlay_instances_mask(img,instances,cmap,alpha=0.9):
    h,w = img.shape[:2]
    overlay = np.zeros((h,w,3),dtype=np.float32)

    _max = instances.max()
    _cmax = cmap.shape[0]
    

    if _max == 0:
        return img
    elif _max > _cmax:
        indexes = [(i % _cmax) for i in range(_max)]    
    else:
        indexes = random.sample(range(0,_cmax),_max)
    
    for i,idx in enumerate(indexes):
        overlay[instances == i+1,:] = cmap[idx,:]
    
    overlay = (overlay * 255.0).astype(np.uint8)
    viz = overlay_rgb_mask(img,overlay,instances>0,alpha=alpha)
    return viz


import pandas as pd
from pandas import ExcelWriter
from pandas import ExcelFile
import re
import numpy as np

def transform_coordinates(B = 0.34527, D = -0.232645, A = 0.0000026948989 , C = 99507.421, F = 3115009.014, file=None):
	
	E = A
	excelFile=pd.read_excel(file)
	#columns = excelFile.columns.tolist()

	# Loop through indices and rows in the dataframe using iterrows
	f = open("BoundingBox.txt","w")
	for index, row in excelFile.iterrows():
		# Loop through columns
		cell = str(row[2])
		name = cell.split("/")[1]
		name = name.split("_")[0]
		# If we find it, print it out
		if name == "20170828bC0970430w280600n" and int(row[0]) == 8035:
				mx = float(row[9])
				my = float(row[8])
				
				a = np.array([[A,B], [D,E]])
				b = np.array([mx - C, my - F])
				Sol1 = np.linalg.solve(a, b)
				Sol1 = [int(round(i)) for i in Sol1]
				
				mx = float(row[7])
				my = float(row[10])
				a = np.array([[A,B], [D,E]])
				b = np.array([mx - C, my - F])
				Sol2 = np.linalg.solve(a, b)
				Sol2 = [int(round(i)) for i in Sol2]
				
				if row[1] == "none":
					f.write("0 ")
				else:
					f.write("1 ")
				f.write(str(Sol1[0]))
				f.write(" ")
				f.write(str(Sol1[1]))
				f.write(" ")
				f.write(str(Sol2[0]))
				f.write(" ")
				f.write(str(Sol2[1]))
				f.write("\n")
				break
	exit()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Dec 26 16:27:06 2020
@author: hasan
"""
import torch
import torch.nn as nn
import torch.nn.functional as F
from typing import Optional, Union, List
from torch.cuda.amp import autocast
#from ..base import modules as md
from segmentation_models_pytorch.base import modules as md
from segmentation_models_pytorch.decoders.unet.decoder import DecoderBlock,CenterBlock
from segmentation_models_pytorch.encoders import get_encoder
from segmentation_models_pytorch.base import SegmentationModel,SegmentationHead, ClassificationHead
from ..models.ASPP import ASPP,DenseASPP
from .RA_modules import Relational_Module
from ..models.OCR import OCR
#from torchviz import make_dot
#8import matplotlib.pyplot as plt
global mod_map,mod_types
mod_map = {
                'aspp' : ASPP,
                'sep_aspp' : ASPP,
                'dense_aspp' : DenseASPP,
                'sep_dense_aspp' : DenseASPP,
                'S_RA_M' : Relational_Module,
                'P_RA_M' : Relational_Module,
                'CR_M' : Relational_Module,
                'SR_M' : Relational_Module,
                'OCR_original' : OCR,
                'OCR_unet' : OCR
                }
mod_types = {
                'aspp' : 'aspp',
                'sep_aspp' : 'aspp',
                'dense_aspp' : 'dense_aspp',
                'sep_dense_aspp' : 'dense_aspp',
                'S_RA_M' : 'relational',
                'P_RA_M' : 'relational',
                'CR_M' : 'relational',
                'SR_M' : 'relational',
                'OCR_original'  : 'context',
                'OCR_unet' : 'context'
                }


class DecoderBlock1x(DecoderBlock):
    
    def __init__(self,
                 in_channels,
                 skip_channels,
                 out_channels,
                 use_batchnorm=True,
                 attention_type=None):
        
        super().__init__(in_channels,
                         skip_channels,
                         out_channels,
                         use_batchnorm,
                         attention_type)
        
    def forward(self, x, skip=None):
        
        if skip is not None:
            x = torch.cat([x, skip], dim=1)
            x = self.attention1(x)
        x = self.conv1(x)
        x = self.conv2(x)
        x = self.attention2(x)
        return x
    
class UnetWMDecoder(nn.Module):
    def __init__(
            self,
            encoder_channels,
            decoder_channels,
            n_blocks=5,
            use_batchnorm=True,
            attention_type=None,
            center=False,
            mod = None,
            mod_configs = {},
            output_stride =32,
            in_size = (512,512)
            
    ):
        super().__init__()
        self.os=output_stride
        if n_blocks != len(decoder_channels):
            raise ValueError(
                "Model depth is {}, but you provide `decoder_channels` for {} blocks.".format(
                    n_blocks, len(decoder_channels)
                )
            )

        encoder_channels = encoder_channels[1:]  # remove first skip with same spatial resolution
        encoder_channels = encoder_channels[::-1]  # reverse channels to start from head of encoder
        #print(encoder_channels)
        
        
        
        if(mod):
            assert mod in mod_map,f'Module {mod} is not implemented!'
            mods = []
            mod_in_channels = encoder_channels[::-1]
            mod_out_channels = []
            for block in range(n_blocks):
                _b = str(block)
                if(_b in mod_configs.keys()):
                    
                    _m = mod_map[mod]
                    _mtype = mod_types[mod]
                    if(_mtype == 'relational'):
                        mod_configs[_b] = {'name' : mod,
                                        'in_channels' : mod_in_channels[block],
                                       'in_size' : self.get_resolution(in_size,block+1)}
                        mods.append(_m(**mod_configs[_b]))
                        mod_out_channels.append(mods[block].out_channels)
                        
                    elif(_mtype == 'aspp'):
                        mod_configs[_b]['separable'] = True if(mod == 'sep_aspp') else False
                        mods.append(_m(mod_in_channels[block],
                                       **mod_configs[_b]))
                        mod_out_channels.append(mod_configs[_b]['out_channels'])
                    elif(_mtype == 'context'):
                        mods.append(_m(mod_in_channels[block],
                                       mod_in_channels[block]//2,
                                       mod_in_channels[block],
                                       0,
                                       1))
                        mod_out_channels.append(mod_in_channels[block])

                    elif(_mtype == 'dense_aspp'):
                        mod_configs[_b]['separable'] = True if(mod == 'sep_dense_aspp') else False
                        mods.append(_m(mod_in_channels[block],
                                           **mod_configs[_b]))
                        #print(mods[-1].out_channels)
                        mod_out_channels.append(mods[-1].out_channels)



                else:
                    mods.append(nn.Identity())
                    mod_out_channels.append(mod_in_channels[block])
                    
            self.with_mods = True
            self.mods = nn.ModuleList(mods[::-1])
            #print(mods)
            
            mod_out_channels = mod_out_channels[::-1]
            #print(mod_out_channels)
            head_channels = mod_out_channels[0]
            in_channels = [head_channels] + list(decoder_channels[:-1])
            skip_channels = list(mod_out_channels[1:]) + [0]
            
        else:
            self.with_mods = False
            self.mods = []
            # computing blocks input and output channels
            head_channels = encoder_channels[0]
            in_channels = [head_channels] + list(decoder_channels[:-1])
            skip_channels = list(encoder_channels[1:]) + [0]
            
        out_channels = decoder_channels
        
        if center:
            self.center = CenterBlock(
                head_channels, head_channels, use_batchnorm=use_batchnorm
            )
        else:
            self.center = nn.Identity()

        # combine decoder keyword arguments
        kwargs = dict(use_batchnorm=use_batchnorm, attention_type=attention_type)
        blocks = [
            DecoderBlock(in_ch, skip_ch, out_ch, **kwargs)
            for in_ch, skip_ch, out_ch in zip(in_channels, skip_channels, out_channels)
        ]
        if(n_blocks == 5 and self.os==16):
            blocks[0]=DecoderBlock1x(in_channels[0],skip_channels[0],out_channels[0])
        self.blocks = nn.ModuleList(blocks)
        
        
                       
                    
                
    def get_resolution(self,in_size = (512,512),stage = 1):
        h,w = in_size
        factor = stage 
        if(self.os == 16 and stage ==5):
            factor -= 1
        return (h//(2**factor),w//(2**factor))
            
        
    def forward(self, *features):

        features = features[1:]    # remove first skip with same spatial resolution
        features = features[::-1]  # reverse channels to start from head of encoder
        if(self.with_mods):
            features  = list(map(lambda x, y : x(y),self.mods,features))
        head = features[0]
        skips = features[1:]

        x = self.center(head)
        for i, decoder_block in enumerate(self.blocks):
            
            skip = skips[i] if i < len(skips) else None
            x = decoder_block(x, skip)

        return x
class UnetWM(SegmentationModel):
 

    def __init__(
        self,
        encoder_name: str = "timm-efficientnet-b0",
        encoder_depth: int = 5,
        encoder_weights: str = None,
        decoder_use_batchnorm: bool = True,
        decoder_channels: List[int] = (256, 128, 64, 32, 16),
        decoder_attention_type: Optional[str] = None,
        in_channels: int = 3,
        classes: int = 1,
        activation: Optional[Union[str, callable]] = None,
        aux_params: Optional[dict] = None,
        mod = None,
        mod_configs = {},
        output_stride = 32,
        in_size = (384,384)
    ):
        super().__init__()

        self.encoder = get_encoder(
            encoder_name,
            in_channels=in_channels,
            depth=encoder_depth,
            weights=encoder_weights,
        )
        if(output_stride == 32):
            pass
        elif(output_stride ==16):
            self.encoder.make_dilated(output_stride)
        else:
            raise ValueError(f'Output stride can only be either 32 or 16 and not {output_stride}')

        self.decoder = UnetWMDecoder(
            encoder_channels=self.encoder.out_channels,
            decoder_channels=decoder_channels,
            n_blocks=encoder_depth,
            use_batchnorm=decoder_use_batchnorm,
            center=True if encoder_name.startswith("vgg") else False,
            attention_type=decoder_attention_type,
            mod = mod,
            mod_configs = mod_configs,
            output_stride = output_stride,
            in_size=in_size
        )

        self.segmentation_head = SegmentationHead(
            in_channels=decoder_channels[-1],
            out_channels=classes,
            activation=activation,
            kernel_size=3,
        )

        if aux_params is not None:
            self.classification_head = ClassificationHead(
                in_channels=self.encoder.out_channels[-1], **aux_params
            )
        else:
            self.classification_head = None

        self.name = "u-{}".format(encoder_name)
        self.initialize()
'''       
_aspp = {
        '4':{
                 'mid_channels' : 256,
                 'inter_channels' : 128,
                 'atrous_rates' : [3,6,12,18],
                 'dropout_rate' : 0,
             }
        }
rr = {'3':{},
      '4' : {}}
print(_aspp)
x = torch.rand((1,3,384,384))
m = UnetWM(mod = 'dense_aspp',
           mod_configs =_aspp)
x = x.cuda(non_blocking = True)
m.train()
m.cuda()
<<<<<<< HEAD
y = m(x) 
=======
with autocast():y = m(x) 
>>>>>>> 9ae615f0dba35685db77246d00157d670e32ded4
print(y.size())
'''



import unittest
from yacs.config import CfgNode as CN
import yaml
import os

class TestConfigManager(unittest.TestCase):
    """
    Test cases for the ConfigManager class methods.
    """

    def setUp(self):
        """
        Set up method to initialize common parameters for tests.
        """
        self.config_manager = ConfigManager(".")

        sample_config = CN()
        sample_config.model = CN()
        sample_config.model.name = "resnet"
        sample_config.model.num_layers = 50
        sample_config.optimizer = CN()
        sample_config.optimizer.type = "adam"
        sample_config.optimizer.lr = 0.001

        # Set the configuration as the cfg attribute of ConfigManager
        self.config_manager.cfg = sample_config

        # Get the configuration dictionary using get_configs method
        retrieved_config = self.config_manager.get_configs()

        # Compare the retrieved configuration with the sample configuration
        self.assertEqual(retrieved_config, sample_config)

    def test_convert_cfg_to_dict(self):
        """
        Test case to verify the convert_cfg_to_dict method of ConfigManager.
        """
        # Define a sample configuration CfgNode
        sample_cfg_node = CN()
        sample_cfg_node.model = CN()
        sample_cfg_node.model.name = "resnet"
        sample_cfg_node.model.num_layers = 50
        sample_cfg_node.optimizer = CN()
        sample_cfg_node.optimizer.type = "adam"
        sample_cfg_node.optimizer.lr = 0.001

        # Convert the sample configuration CfgNode to a dictionary
        converted_dict = self.config_manager.convert_cfg_to_dict(sample_cfg_node)

        # Define the expected dictionary
        expected_dict = {
            "model": {"name": "resnet", "num_layers": 50},
            "optimizer": {"type": "adam", "lr": 0.001}
        }

        # Compare the converted dictionary with the expected dictionary
        self.assertEqual(converted_dict, expected_dict)

    def test_save_cfg_as_yaml(self):
        """
        Test case to verify the save_cfg_as_yaml method of ConfigManager.
        """
        # Define a sample configuration dictionary
        sample_config = {
            "model": {"name": "resnet", "num_layers": 50},
            "optimizer": {"type": "adam", "lr": 0.001}
        }

        # Call the save_cfg_as_yaml method with the sample configuration dictionary
        self.config_manager.save_cfg_as_yaml(sample_config)

        # Check if the YAML file has been created
        yaml_file_path = f"{self.config_manager.config_file_path}/configs.yaml"
        self.assertTrue(os.path.exists(yaml_file_path))

        # Read the saved YAML file and check if it matches the sample configuration
        with open(yaml_file_path, "r") as yaml_file:
            saved_config = yaml.safe_load(yaml_file)
        self.assertEqual(saved_config, sample_config)

    def test_load_cfg_from_yaml(self):
        """
        Test case to verify the _load_cfg_from_yaml method of ConfigManager.
        """
        # Define a sample configuration dictionary
        sample_config = {
            "model": {"name": "resnet", "num_layers": 50},
            "optimizer": {"type": "adam", "lr": 0.001}
        }

        # Write the sample configuration to a YAML file
        yaml_file_path = f"{self.config_manager.config_file_path}/configs.yaml"
        with open(yaml_file_path, "w") as yaml_file:
            yaml.safe_dump(sample_config, yaml_file, default_flow_style=False)

        # Call the _load_cfg_from_yaml method
        loaded_config_node = self.config_manager._load_cfg_from_yaml()

        # Convert the loaded CfgNode instance to a dictionary
        loaded_config_dict = self.config_manager.convert_cfg_to_dict(loaded_config_node)

        # Check if the loaded configuration matches the sample configuration
        self.assertEqual(loaded_config_dict, sample_config)



import numpy as np
import pandas as pd
import geopandas as gpd
import rasterio as rio
from rasterio.features import shapes
from skimage.morphology import dilation,square,erosion
from skimage.segmentation import watershed
from skimage.measure import label
from PIL import Image,ImageDraw
import pandas as pd
from shapely.geometry import shape
from shapely.wkt import dumps
from shapely.ops import unary_union


class PostProcessing():
    def post_process(self, pred, thresh=0.5, thresh_b=0.6, mina=100, mina_b=50):
        """
        Post-processes a prediction mask to obtain a refined segmentation.This function refines 
        a semantic segmentation mask, particularly for building segmentation tasks.
        It leverages optional channels for predicting borders and spacing around buildings.

        Args:
            pred (numpy.ndarray): Prediction mask with shape (height, width, channels).
            thresh (float, default=0.5): Threshold for considering pixels as part of the final segmentation.
            thresh_b (float, default=0.6): Threshold for considering pixels as borders between objects.
            mina (int, default=100): Minimum area threshold for retaining segmented regions.
            mina_b (int, default=50): Minimum area threshold for retaining basins.

        Returns:
            numpy.ndarray: Refined segmentation mask.

        Description:
            The refinement process involves:
            1. Extracting individual channels from the input mask.
            2. Separating nuclei (building interiors) from borders based on the predicted borders channel.
            3. Applying thresholding to identify basins within nuclei, which represent potential individual buildings.
            4. Optionally filtering out small basins based on the minimum area threshold.
            5. Performing watershed segmentation to separate closely located buildings.
            6. Applying noise filtering to remove small regions from the segmented mask.
            7. Returning the refined segmentation mask with labeled and filtered regions.

        Note:
            - The function assumes a specific order for input channels:
                - Channel 0: Building predictions
                - Channel 1 (optional): Border predictions
                - Channel 2 (optional): Spacing predictions
            - The output represents labeled regions in the refined segmentation.

        """
        if len(pred.shape) < 2:
            return None
        if len(pred.shape) == 2:
            pred = pred[..., np.newaxis]

        ch = pred.shape[2]
        buildings = pred[..., 0]

        if ch > 1:
            borders = pred[..., 1]
            nuclei = buildings * (1.0 - borders)

            if ch == 3:
                spacing = pred[..., 2]
                nuclei *= (1.0 - spacing)

            basins = label(nuclei > thresh_b, background=0, connectivity=2)

            if mina_b > 0:
                basins = self.noise_filter(basins, mina=mina_b)
                basins = label(basins, background=0, connectivity=2)

            washed = watershed(image=-buildings, markers=basins, mask=buildings > thresh, watershed_line=False)

        elif ch == 1:
            washed = buildings > thresh

        washed = label(washed, background=0, connectivity=2)
        washed = self.noise_filter(washed, mina=mina)
        washed = label(washed, background=0, connectivity=2)

        return washed
    
    @staticmethod
    def noise_filter(washed,mina):
        """
        Filter small regions in a labeled segmentation mask based on minimum area.
        This function filters out small labeled regions in a segmentation mask based on their area.
        It iterates over unique labels, calculates the area for each label, and sets the pixel values
        corresponding to labels with area less than or equal to the specified threshold to 0.

        Args:
            washed (numpy.ndarray): Input labeled segmentation mask.
            mina (int): Minimum area threshold for retaining labeled regions.

        Returns:
            numpy.ndarray: Segmentation mask with small labeled regions filtered out.
        """
        values = np.unique(washed)
        for val in values[1:]:
            area = (washed[washed == val]>0).sum()
            if area<=mina:
                washed[washed == val] = 0
        return washed

    @staticmethod
    def extract_poly(mask):
        """
        Extract polygons from a binary mask using the `rasterio.features.shapes` method.
        It then processes each shape, converts it to a Shapely Polygon, and appends it to a list.
        The list of polygons is then combined into a single geometry using `shapely.ops.unary_union`.
        The resulting polygon or None (if no polygons are found) is returned.

        Args:
            mask (numpy.ndarray): Binary input mask.

        Returns:
            shapely.geometry.base.BaseGeometry or None: Resulting polygon or None if no polygons are found.
        """
        shps = shapes(mask.astype(np.int16), mask > 0)
        polys = []
        #check for validity to avoid crashes
        try:
            for shp, _ in shps:
                p = shape(shp).buffer(0.0)
                typ = p.geom_type
                if typ == 'Polygon' or typ == 'MultiPolygon':
                    polys.append(p.simplify(0.01))
                else:
                    continue
        except:
            return None

        if len(polys) == 0:
            return None
        else:
            return unary_union(polys)
        
    @staticmethod    
    def instance_mask_to_gdf(instance_mask, transform=None, crs=None):
        """
        Convert an instance mask to a GeoDataFrame, uses rasterio's `shapes` method to obtain shapes from the instance mask.
        The shapes are then converted to a GeoDataFrame, where each geometry is associated with a unique id.
        The resulting GeoDataFrame is dissolved by the 'id' column to form a single geometry for each instance.
        If no instances are found, an empty GeoDataFrame is returned.

        Args:
            instance_mask (numpy.ndarray): Input instance mask with shape (H, W), where each instance is labeled by a unique id/number.
            transform (affine.Affine, Optional): Geospatial transform of the raster. Default is None.
            crs (str, Optional): CRS of the raster. Default is None.

        Returns:
            geopandas.GeoDataFrame: GeoDataFrame of the shapes projected to the specified CRS using the transform.
        """
        # Transform should be Identity if None is provided
        transform = rio.transform.IDENTITY if transform is None else transform

        all_shapes = shapes(instance_mask, mask=None, transform=transform)
        data = [{'properties': {'id': v}, 'geometry': s} for i, (s, v) in enumerate(all_shapes) if v != 0]

        if len(data) == 0:
            # Return empty GeoDataFrame
            return gpd.GeoDataFrame(columns=['id', 'geometry'], geometry='geometry', crs=crs)

        gdf = gpd.GeoDataFrame.from_features(data, crs=crs)
        gdf = gdf.dissolve(by='id')

        return gdf



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Mar 11 18:30:12 2020

@author: jamada
"""

#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Mar 11 16:17:07 2020

@author: hasan
"""
""" at some point i tested a 2 classes segmentation with classes=['buildings','spacing']
    so i made this script to modify the spacing....i tried at first the spacing of really close buildings only labelled, 
    but then we decided to test a bigger spacing width...which means to increase the range to label the spacing of buildings,
    u can see the changes in the make_separation function between here and make_dataset.py script

most of these functions were commented before in make_dataset.py
"""
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)
from skimage.morphology import dilation, square
from skimage.segmentation import watershed
import cv2
import os
import argparse
from tqdm import tqdm
import numpy as np
from PIL import Image,ImageDraw
from imantics import Mask
from simplification.cutil import simplify_coords_vwp
def create_dir(path):
    try:
        os.mkdir(path)
    except OSError:
        print ("Creation of the directory %s failed" % path)
def create_separation(labels):
    tmp = dilation(labels > 0, square(23))
    tmp2 = watershed(tmp, labels, mask=tmp, watershed_line=True) > 0
    tmp = tmp ^ tmp2
    tmp = dilation(tmp, square(11))
    msk1 = np.zeros_like(labels, dtype='bool')

    for y0 in range(labels.shape[0]):
        for x0 in range(labels.shape[1]):
            if not tmp[y0, x0]:
                continue
            if labels[y0, x0] == 0:
                sz = 17
            else:
                sz = 23
            uniq = np.unique(labels[max(0, y0 - sz):min(labels.shape[0], y0 + sz + 1),
                             max(0, x0 - sz):min(labels.shape[1], x0 + sz + 1)])
            if len(uniq[uniq > 0]) > 1:
                msk1[y0, x0] = True
    return msk1

def shape_polys(polyg):
    all_polys = []
    for poly in polyg:
        if len(poly) >= 3:
            f = poly.reshape(-1, 2)
            simplified_vw = simplify_coords_vwp(f, .3)
            if len(simplified_vw) > 2:
                mpoly = []
                # Rebuilding the polygon in the way that PIL expects the values [(x1,y1),(x2,y2)]
                for i in simplified_vw:
                    mpoly.append((i[0], i[1]))
                    # Adding the first point to the last to close the polygon
                mpoly.append((simplified_vw[0][0], simplified_vw[0][1]))
                all_polys.append(mpoly)
    return all_polys

def make_instance_mask(all_polys,size):
    bg=np.zeros((size,size)).astype(np.uint8)
    bg=Image.fromarray(bg).convert('L')
    shift=255-len(all_polys)
    for i,poly in enumerate(all_polys):
        ImageDraw.Draw(bg).polygon(poly,outline=shift+i,fill=shift+i)
    return bg

def new_spacing(mask):
    #get the old mask, where buildings are 64 encoded,borders are 128, and spacing is 255
    w,h=mask.shape
    #remove the old spacing, set to zero
    mask[(np.where(mask==np.uint8(255)))]=np.uint8(0)
    #get the polygons of the buildings (64)
    polyg = Mask(mask==np.uint8(64)).polygons()
    #rest is explained in make_dataset.py
    extracted=shape_polys(polyg)
    labels=np.array(make_instance_mask(extracted,w))
    spacing=np.array(create_separation(labels)).astype(np.uint8)*255
    spacing[(np.where(mask==np.uint8(128)))] = np.uint8(128)
    spacing[(np.where(mask==np.uint8(64)))] = np.uint8(64)
    return spacing
def main(old,new):
    paths=[f for f in os.listdir(old)]
    #iterate over the old dataset masks and make new masks with modified spacing
    for pth in tqdm(paths):
        mask=cv2.imread(f'{old}/{pth}',0)
        new_mask=new_spacing(mask=mask)
        Image.fromarray(new_mask).save(f'{new}/{pth}')
    print('DOne!')

if __name__ == '__main__':
    parser=argparse.ArgumentParser()
    parser.add_argument('--old',type=str)
    parser.add_argument('--new',type=str)
    args=parser.parse_args()
    create_dir(args.new)
    main(old=args.old,new=args.new)
    

      













    import numpy as np
    from tmp import get_micro_metrics

    class Meter(object):
        """
        Meter is a general interface / class to use as a base for our classes
        With no implemented methods (to be implemented in children)
        """
        def reset(self):
            """
            Resets the meter to its initial state.
            """

        def update(self,value):
            """
            Updates the meter with a new value.

            Args:
                value: The value to update the meter with.
            """

        def get_update(self):
            """
            Gets the current state or value of the meter.

            Returns:
                The current state or value of the meter.
            """

    class AverageMeter(Meter):
        """
        Initializes an instance of the AverageMeter class subclass of Meter.
        """
        def __init__(self):
            super(AverageMeter,self).__init__()
            self.value=0.
            self.average=0.
            self.count=0.

        def reset(self):
            """
            Resets the AverageMeter to its initial state.
            """
            self.value = 0.
            self.average = 0.
            self.count = 0.
            
        def update(self,value):
            """
            Updates the AverageMeter with a new value.

            Args:
                value: The value to update the AverageMeter with.
            """
            self.count += 1.
            self.value = value
            self.average = (self.average * (self.count - 1)+ self.value)/float(self.count)
            
        def get_update(self):
            """
            Gets the current average value of the AverageMeter.

            Returns:
                The current average value of the AverageMeter.
            """
            return self.average

    class ShortTermMemoryMeter(Meter):
        """
        Initializes an instance of the ShortTermMemoryMeter class sublcass of Meter.

        Args:
            memory_length (int): Length of the short-term memory window.
                                Should be greater than 1 for meaningful averaging.
        """
        def __init__(self,memory_length):
            super(ShortTermMemoryMeter,self).__init__()
            self.value = 0
            self.length = 0
            self.in_memory = []
            self.average = 0
            self.memory_length = memory_length
            assert(self.memory_length >1)
            #to at least average last 2 batches results or else it would be only performing additional non-neccessary operations 
            
        def reset(self):
            """
            Resets the ShortTermMemoryMeter to its initial state.
            """
            self.value = 0
            self.length = 0
            self.in_memory = []
            self.average = 0
            
        def update(self,value):
            """
            Updates the ShortTermMemoryMeter with a new value.

            Args:
                value: The value to update the ShortTermMemoryMeter with.
            """
            self.value = value
            if (self.length >= self.memory_length):
                self.in_memory = self.in_memory[1:]
            self.in_memory.append(self.value)
            self.average = np.average(np.array(self.in_memory))
            self.length = len(self.in_memory)
            
        def get_update(self):
            """
            Gets the current average value of the ShortTermMemoryMeter.

            Returns:
                The current average value of the ShortTermMemoryMeter.
            """
            return self.average

    class Sampler(Meter):
        """
        Initializes an instance of the Sampler class.

        Args:
            value_names (list): List of names for the sampled values.
            rate (int): Sampling rate, i.e., how often to store values in the history.
        """
        def __init__(self,value_names,rate):
            super(Sampler,self).__init__()
            self.v_names = value_names
            self.rate = rate
            self.reset()
            
        def reset(self):
            """
            Resets the Sampler to its initial state.
            """
            self.history=dict()
            for name in self.v_names:
                self.history[name] = list()
        
        def update(self,values,i):
            """
            Updates the Sampler with a new set of values at a given iteration.

            Args:
                values (list): List of values to update the Sampler with.
                i (int): Current iteration or step.

            Note:
                - Values are stored in the history based on the specified sampling rate.
            """
            if(i%self.rate == 0):
                for j,k in enumerate(values):
                    name = self.v_names[j]
                    self.history[name].append(k)
                    
        def get_update(self):
            """
            Gets the current history of sampled values.

            Returns:
                dict: Dictionary containing sampled values for each value name.
            """
            return self.history

    class MetricsManager():
        """
        A class for managing and reporting various metrics during training or validation.
        """ 
        classes = ['overall','background' , 'building_flooded' ,'building_non-flooded' , 'road_flooded' ,
            'road_non-flooded' , 'water' , 'tree' , 'vehicle' , 'pool' ,  'grass']
        
        def __init__(self, classes, metric_names=['iou'], thresh=0.5):
            """
            Initializes an instance of the MetricsManager class.

            Args:
                classes (list): List of class names or labels.
                metric_names (list, optional): List of metric names. Default is ['iou'].
                thresh (float, optional): Threshold value for metrics. Default is 0.5.
            """
            self.classes = classes
            self.metric_names = metric_names
            self.thresh = thresh
            self.metrics = self.register_metrics_avg_meters()
            
        def register_metrics_avg_meters(self):
            """
            Registers metric functions with AverageMeter instances.

            Returns:
                list: List of tuples containing metric functions and corresponding AverageMeter instances.
            """
            l = len(self.classes)
            #get_micro_metrics is from the NoteBook which includes a lot of dependecies -> new file with all functions ?
            metrics = get_micro_metrics(
                metrics=[self.metric_names] * l,
                threshs=[self.thresh] * l,
                channels=[None,*[i for i in range(l-1)]],
                names=self.classes,
                num_cls=l
            )
            tups = []
            for metric in metrics:
                tups.append([metric,AverageMeter()])
            return tups

        @staticmethod
        def get_meters_info(seg_loss_meter,seg_loss_name,cls_loss_meter,cls_loss_name, mnms,cls_accuracy_meter,cls_f1score_meter):
            """
            Gets a formatted string with information about various metrics.

            Args:
                seg_loss_meter (Meter): Meter instance for segmentation loss.
                seg_loss_name (str): Name for segmentation loss.
                cls_loss_meter (Meter): Meter instance for classification loss.
                cls_loss_name (str): Name for classification loss.
                mnms (list): List of tuples containing metric functions and corresponding Meter instances.
                cls_accuracy_meter (Meter): Meter instance for classification accuracy.
                cls_f1score_meter (Meter): Meter instance for classification F1 score.

            Returns:
                str: Formatted string with information about various metrics.
            """
            info = ''
            info += f'| {seg_loss_name} : {seg_loss_meter.get_update():.5} '
            info += f'| {cls_loss_name} : {cls_loss_meter.get_update():.5} '
            info += f'| cls_accuracy : {cls_accuracy_meter.get_update():.5} '
            info += f'| cls_f1score : {cls_f1score_meter.get_update():.5} '
            for metric,meter in mnms:
                info += f'| {metric.__name__} : {meter.get_update():.5}'
            info += ' |'
            return info

        @staticmethod
        def register_scores(pdict,seg_loss_meter,seg_criterion,cls_loss_meter,cls_criterion, mnms,cls_accuracy_meter,cls_f1score_meter,prefix = 'train'):
            """
            Registers scores for different metrics in a provided dictionary.

            Args:
                pdict (dict): Dictionary to store metric scores.
                seg_loss_meter (Meter): Meter instance for segmentation loss.
                seg_criterion (torch.nn.Module): Segmentation loss criterion.
                cls_loss_meter (Meter): Meter instance for classification loss.
                cls_criterion (torch.nn.Module): Classification loss criterion.
                mnms (list): List of tuples containing custom metrics and their corresponding Meter instances.
                cls_accuracy_meter (Meter): Meter instance for classification accuracy.
                cls_f1score_meter (Meter): Meter instance for classification F1 score.
                prefix (str, optional): Prefix for metric names (e.g., 'train' or 'val'). Default is 'train'.

            Returns:
                dict: Updated dictionary with metric scores.
            """
            assert prefix in ['train','val'],'{} is not a valid prefix'.format(prefix)
            seg_loss_name = f'{prefix}_seg_' + seg_criterion.__name__
            cls_loss_name = f'{prefix}_cls_' + cls_criterion.__name__
            keys = list(pdict.keys())

            if(seg_loss_name not in keys):
                pdict[seg_loss_name] = []
                pdict[cls_loss_name] = []
                pdict['cls_accuracy'] = []
                pdict['cls_f1score'] = []

            pdict[seg_loss_name].append(seg_loss_meter.get_update())
            pdict[cls_loss_name].append(cls_loss_meter.get_update())
            pdict['cls_accuracy'].append(cls_accuracy_meter.get_update())
            pdict['cls_f1score'].append(cls_f1score_meter.get_update())

            for metric,meter in mnms:
                if(f'{prefix}_' + metric.__name__ not in keys):
                    pdict[f'{prefix}_' + metric.__name__] = []
                pdict[f'{prefix}_' + metric.__name__].append(meter.get_update())
            return pdict



import unittest
import torch
from torch import nn
from optimizers import get_optimizer
from schedulers import AutoScheduler,scheduler_mapping,get_scheduler
import numpy as np


class testScheduler(unittest.TestCase):
    def test_keys_lower_case(self):
        for schechduler in scheduler_mapping:
            self.assertEqual(schechduler,schechduler.lower())



if __name__=="__main__":
    unittest.main()



import sys
import gc
class SAM:
   
    def __init__(self,checkpoint):
        import sys
        #sys.path.append("..")
        from segment_anything import sam_model_registry
        model_type = "vit_h"
        device = "cuda"
        self.sam = sam_model_registry[model_type](checkpoint=checkpoint)
        self.sam.to(device=device)

    def predictSAM(self,x,image,input_point=None,input_label=None,input_boxes=None,mask_input=None,flag=0):
        if flag==1:
                output = self.sam.forward(
                batched_input=[
                    {
                        "image": x,
                        "original_size": image.shape[:2],
                        'boxes':input_boxes,     
                    }
                ],
                multimask_output=False,
                )[0]
        else :
            output = self.sam.forward(
                batched_input=[
                    {
                        "image": x,
                        "original_size": image.shape[:2],
                        'point_coords':input_point,
                        'point_labels':input_label,
                        'boxes':input_boxes
                        # 'mask_inputs':mask_input
                    }
                ],
                multimask_output=False,
            )[0]

        ##Uncomment to select the mask of max iou in case of multimask output is true
        # pred_mask=[]
        # for mask,score in zip(output["masks"],output["iou_predictions"]):
        #   max_score=torch.argmax(score)
        #   pred_mask.append(mask[max_score])
        # pred_mask=torch.stack(pred_mask)
        # pred_mask=pred_mask.unsqueeze(1).cuda()

        ##Uncomment for multimask output which take the last mask
        # pred_mask=[]
        # for mask in output["masks"]:
        #   pred_mask.append(mask[2])
        # pred_mask=torch.stack(pred_mask)
        # pred_mask=pred_mask.unsqueeze(1).cuda()

        pred_mask = output["masks"]
        gc.collect()
        del output
        return pred_mask


import unittest
from unittest.mock import patch
import os
import json
import numpy as np
import pandas as pd
import glob
from matplotlib import pyplot as plt
import geopandas as gpd
from shapely.geometry import Polygon

class TestIoUNumpy(unittest.TestCase):
    def test_iou_numpy_perfect_match(self):
        outputs = torch.tensor([[1, 0, 1], [1, 1, 0], [0, 1, 1]])
        labels = torch.tensor([[1, 0, 1], [1, 1, 0], [0, 1, 1]])
        result = iou_numpy(outputs, labels)
        self.assertEqual(result, 1.0, "IoU should be 1.0 for perfect match")

    def test_iou_numpy_no_overlap(self):
        outputs = torch.tensor([[0, 0, 0], [0, 0, 0], [0, 0, 0]])
        labels = torch.tensor([[1, 1, 1], [1, 1, 1], [1, 1, 1]])
        result = iou_numpy(outputs, labels)
        self.assertEqual(result, 0.0, "IoU should be 0.0 for no overlap")

    def test_iou_numpy_empty_input(self):
        outputs = torch.tensor([])
        labels = torch.tensor([])
        result = iou_numpy(outputs, labels)
        self.assertTrue(torch.isnan(result), "IoU should be NaN for empty input")
        
class TestMatchingAlgorithm(unittest.TestCase):
    def setUp(self):
        # Sample ground truth and predicted bounding boxes for testing
        self.gt_bbox = [np.array([[1, 1, 3, 3], [4, 4, 6, 6]]),
                        np.array([[7, 7, 9, 9], [10, 10, 12, 12]])]
        self.pred_bbox = [np.array([[0, 0, 2, 2], [2, 2, 5, 5]]),
                          np.array([[7, 7, 9, 9], [9, 9, 12, 12]])]

    def test_matching_method(self):
        matcher = MatchingAlgorithm(self.gt_bbox, self.pred_bbox)
        iou_list, f1_scores, tp_pred_indices, tp_gt_indices, fp_indices, fn_indices, _, precision, recall = matcher.matching()

        self.assertEqual(len(iou_list), 2, "Incorrect number of IoU scores")
        self.assertEqual(len(f1_scores), 2, "Incorrect number of F1 scores")
        self.assertEqual(len(tp_pred_indices), 2, "Incorrect number of true positive predicted indices")
        self.assertEqual(len(tp_gt_indices), 2, "Incorrect number of true positive ground truth indices")
        self.assertEqual(len(fp_indices), 0, "Incorrect number of false positive indices")
        self.assertEqual(len(fn_indices), 0, "Incorrect number of false negative indices")
        self.assertAlmostEqual(precision, 1.0, places=2, msg="Incorrect precision value")
        self.assertAlmostEqual(recall, 1.0, places=2, msg="Incorrect recall value")

    def test_tp_iou_method(self):
        matcher = MatchingAlgorithm(self.gt_bbox, self.pred_bbox)
        _, _, tp_pred_indices, tp_gt_indices, _, _, _, _, _ = matcher.matching()
        tp_iou_list, avg_tp_iou = matcher.tp_iou(tp_pred_indices, tp_gt_indices)
        self.assertEqual(len(tp_iou_list), 2, "Incorrect number of true positive IoU scores")
        self.assertAlmostEqual(avg_tp_iou, 0.875, places=2, msg="Incorrect average true positive IoU value")

class TestCalScores(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.output_dir = 'test_output'
        cls.score_dir = 'test_scores'
        os.makedirs(cls.output_dir, exist_ok=True)
        os.makedirs(cls.score_dir, exist_ok=True)

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.output_dir)
        shutil.rmtree(cls.score_dir)

    def test_initialization(self):
        scores = CalScores(self.output_dir, self.score_dir)
        self.assertEqual(scores.output_dir, self.output_dir)
        self.assertEqual(scores.score_dir, self.score_dir)
        
    @patch("builtins.open")
    def test_micro_match_iou(self, mock_open):
        # Test inputs
        output_dir = "output"
        score_dir = "scores"
        pred_mask = torch.zeros(2, 1, 10, 10)
        name = "test_image"
        gt = {"geometry": []}
        image = np.zeros((10, 10, 3))
        input_point = None
        input_label = None
        tile_boxes = [(1, 1, 5, 5)]

        # Initialize CalScores object
        cal_scores = CalScores(output_dir, score_dir)
        cal_scores.micro_match_iou(pred_mask, name, gt, image, input_point, input_label, tile_boxes, save=None, visualize=True)


    def test_macro_score(self):
        output_dir = "output"
        score_dir = "scores"
        cal_scores = CalScores(output_dir, score_dir)

        with patch("os.path.join"), patch("os.listdir"), patch("glob.glob"), patch("pandas.DataFrame.to_csv"), patch("builtins.open", create=True) as mocked_open, patch("json.dump"):
            cal_scores.macro_score()




import math
import torch
from torch.optim.optimizer import Optimizer
import numpy as np
import torch.nn as nn

# Original source:  https://github.com/shivram1987/diffGrad/blob/master/diffGrad.py

# modifications: @lessw2020
# https://github.com/lessw2020/Best-Deep-Learning-Optimizers/blob/master/diffgrad/diffgrad.py

class DiffGrad(Optimizer):
    r"""Implements diffGrad algorithm. It is modified from the pytorch implementation of Adam.
    It has been proposed in `diffGrad: An Optimization Method for Convolutional Neural Networks`_.
    Arguments:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float, optional): learning rate (default: 1e-3)
        betas (Tuple[float, float], optional): coefficients used for computing
            running averages of gradient and its square (default: (0.9, 0.999))
        eps (float, optional): term added to the denominator to improve
            numerical stability (default: 1e-8)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
        amsgrad (boolean, optional): whether to use the AMSGrad variant of this
            algorithm from the paper `On the Convergence of Adam and Beyond`_
            (default: False)
    .. _diffGrad: An Optimization Method for Convolutional Neural Networks:
        https://arxiv.org/abs/1909.11015
    .. _Adam\: A Method for Stochastic Optimization:
        https://arxiv.org/abs/1412.6980
    .. _On the Convergence of Adam and Beyond:
        https://openreview.net/forum?id=ryQu7f-RZ
    """

    def __init__(self, params, lr=1e-3, betas=(0.9, 0.999), eps=1e-8, version=0, weight_decay=0):
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))
        
        
        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay)
        
        super().__init__(params, defaults)
        
        #save version
        self.version = version

    def __setstate__(self, state):
        super().__setstate__(state)

    def step(self, closure=None):
        """Performs a single optimization step.
        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError('diffGrad does not support sparse gradients, please consider SparseAdam instead')

                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state['step'] = 0
                    # Exponential moving average of gradient values
                    state['exp_avg'] = torch.zeros_like(p.data)
                    # Exponential moving average of squared gradient values
                    state['exp_avg_sq'] = torch.zeros_like(p.data)
                    # Previous gradient
                    state['previous_grad'] = torch.zeros_like(p.data)

                exp_avg, exp_avg_sq, previous_grad = state['exp_avg'], state['exp_avg_sq'], state['previous_grad']
                beta1, beta2 = group['betas']

                state['step'] += 1

                if group['weight_decay'] != 0:
                    grad.add_(group['weight_decay'], p.data)

                # Decay the first and second moment running average coefficient
                exp_avg.mul_(beta1).add_(1 - beta1, grad)
                exp_avg_sq.mul_(beta2).addcmul_(1 - beta2, grad, grad)
                denom = exp_avg_sq.sqrt().add_(group['eps'])

                bias_correction1 = 1 - beta1 ** state['step']
                bias_correction2 = 1 - beta2 ** state['step']

                # compute diffgrad coefficient (dfc)
                
                
                if self.version==0:
                    diff = abs(previous_grad - grad)
                elif self.version ==1:
                    diff = previous_grad-grad
                elif self.version ==2:
                    diff =  .5*abs(previous_grad - grad)
                    
                if self.version==0 or self.version==1:    
                    dfc = 1. / (1. + torch.exp(-diff))
                elif self.version==2:
                    dfc = 9. / (1. + torch.exp(-diff))-4      #DFC2 = 9/(1+e-(.5/g/)-4 #range .5,5
                    
                state['previous_grad'] = grad

                # update momentum with dfc
                exp_avg1 = exp_avg * dfc

                step_size = group['lr'] * math.sqrt(bias_correction2) / bias_correction1

                p.data.addcdiv_(-step_size, exp_avg1, denom)

        return loss


from torch.optim import Optimizer,Adam,RMSprop,SGD

from .over9000 import *
from .adamw import AdamW

optimizer_mapping = {
    'adam' : Adam,
    'adamw' : AdamW,
    'rmsprop' : RMSprop,
    'sgd' : SGD,
    'radam' : RAdam,
    'plain_radam' : PlainRAdam,
    'novograd' : Novograd,
    'ranger' : Ranger,
    'ralamb' : Ralamb,
    ################
    'rangerlars' : RangerLars, ##Known as Over9000 optimizer
    'over9000' : RangerLars,
    ################
        # 'lookahead' : Lookahead,
    'lookahead_adam' : LookaheadAdam,
    'diffgrad' : DiffGrad,
    'adamod' : AdaMod,
    'madam' : Madam,
    'apollo' : Apollo,
    'adabelief' : AdaBelief,
    'madgrad' : MADGRAD,
    'adan' : Adan

}

def get_optimizer(name:str ,params,lookAhead=False,lookAhead_alpha=0.5, lookAhead_k=6,lookAhead_pullback_momentum="none",*args,**kwargs) -> Optimizer:
    """
    This function returns the optimizer given its name

    @param name: name of the optimzer
    @param params: parameters of the model that need to be optimzed 
    @param *args & **kwargs: parameters for the optimizer

    @type name:str
    @type params: list or dict
    
    @return: torch.optim.Optimzer

    """
    name = name.lower()
    if name not in optimizer_mapping.keys():
        raise ValueError('Optimizer {} not an option'.format(name))

    if lookAhead:
        assert name != 'adan', "lookahead adan is not supported"
        return Lookahead(optimizer_mapping[name](params,*args,**kwargs),alpha=lookAhead_alpha, k=lookAhead_k,pullback_momentum=lookAhead_pullback_momentum)
    return optimizer_mapping[name](params,*args,**kwargs)


"""
Lovasz-Softmax and Jaccard hinge loss in PyTorch
Maxim Berman 2018 ESAT-PSI KU Leuven (MIT License)
"""

from __future__ import print_function, division

import torch
import torch.nn.functional as F
from torch.autograd import Variable
from torch.nn.modules.loss import _Loss

try:
    from itertools import ifilterfalse
except ImportError:  # py3k
    from itertools import filterfalse as ifilterfalse

__all__ = ["BinaryLovaszLoss", "LovaszLoss"]


def _lovasz_grad(gt_sorted):
    """Compute gradient of the Lovasz extension w.r.t sorted errors
    See Alg. 1 in paper
    """
    p = len(gt_sorted)
    gts = gt_sorted.sum()
    intersection = gts - gt_sorted.float().cumsum(0)
    union = gts + (1 - gt_sorted).float().cumsum(0)
    jaccard = 1.0 - intersection / union
    if p > 1:  # cover 1-pixel case
        jaccard[1:p] = jaccard[1:p] - jaccard[0:-1]
    return jaccard


def _lovasz_hinge(logits, labels, per_image=True, ignore=None):
    """
    Binary Lovasz hinge loss
        logits: [B, H, W] Variable, logits at each pixel (between -infinity and +infinity)
        labels: [B, H, W] Tensor, binary ground truth masks (0 or 1)
        per_image: compute the loss per image instead of per batch
        ignore: void class id
    """
    if per_image:
        loss = mean(
            _lovasz_hinge_flat(*_flatten_binary_scores(log.unsqueeze(0), lab.unsqueeze(0), ignore))
            for log, lab in zip(logits, labels)
        )
    else:
        loss = _lovasz_hinge_flat(*_flatten_binary_scores(logits, labels, ignore))
    return loss


def _lovasz_hinge_flat(logits, labels):
    """Binary Lovasz hinge loss
    Args:
        logits: [P] Variable, logits at each prediction (between -iinfinity and +iinfinity)
        labels: [P] Tensor, binary ground truth labels (0 or 1)
        ignore: label to ignore
    """
    if len(labels) == 0:
        # only void pixels, the gradients should be 0
        return logits.sum() * 0.0
    signs = 2.0 * labels.float() - 1.0
    errors = 1.0 - logits * Variable(signs)
    errors_sorted, perm = torch.sort(errors, dim=0, descending=True)
    perm = perm.data
    gt_sorted = labels[perm]
    grad = _lovasz_grad(gt_sorted)
    loss = torch.dot(F.relu(errors_sorted), Variable(grad))
    return loss


def _flatten_binary_scores(scores, labels, ignore=None):
    """Flattens predictions in the batch (binary case)
    Remove labels equal to 'ignore'
    """
    scores = scores.view(-1)
    labels = labels.view(-1)
    if ignore is None:
        return scores, labels
    valid = labels != ignore
    vscores = scores[valid]
    vlabels = labels[valid]
    return vscores, vlabels


# --------------------------- MULTICLASS LOSSES ---------------------------


def _lovasz_softmax(probas, labels, classes="present", per_image=False, ignore=None):
    """Multi-class Lovasz-Softmax loss
    Args:
        @param probas: [B, C, H, W] Variable, class probabilities at each prediction (between 0 and 1).
        Interpreted as binary (sigmoid) output with outputs of size [B, H, W].
        @param labels: [B, H, W] Tensor, ground truth labels (between 0 and C - 1)
        @param classes: 'all' for all, 'present' for classes present in labels, or a list of classes to average.
        @param per_image: compute the loss per image instead of per batch
        @param ignore: void class labels
    """
    if per_image:
        loss = mean(
            _lovasz_softmax_flat(*_flatten_probas(prob.unsqueeze(0), lab.unsqueeze(0), ignore), classes=classes)
            for prob, lab in zip(probas, labels)
        )
    else:
        loss = _lovasz_softmax_flat(*_flatten_probas(probas, labels, ignore), classes=classes)
    return loss


def _lovasz_softmax_flat(probas, labels, classes="present"):
    """Multi-class Lovasz-Softmax loss
    Args:
        @param probas: [P, C] Variable, class probabilities at each prediction (between 0 and 1)
        @param labels: [P] Tensor, ground truth labels (between 0 and C - 1)
        @param classes: 'all' for all, 'present' for classes present in labels, or a list of classes to average.
    """
    if probas.numel() == 0:
        # only void pixels, the gradients should be 0
        return probas * 0.0
    C = probas.size(1)
    losses = []
    class_to_sum = list(range(C)) if classes in ["all", "present"] else classes
    for c in class_to_sum:
        fg = (labels == c).float()  # foreground for class c
        if classes == "present" and fg.sum() == 0:
            continue
        if C == 1:
            if len(classes) > 1:
                raise ValueError("Sigmoid output possible only with 1 class")
            class_pred = probas[:, 0]
        else:
            class_pred = probas[:, c]
        errors = (Variable(fg) - class_pred).abs()
        errors_sorted, perm = torch.sort(errors, 0, descending=True)
        perm = perm.data
        fg_sorted = fg[perm]
        losses.append(torch.dot(errors_sorted, Variable(_lovasz_grad(fg_sorted))))
    return mean(losses)


def _flatten_probas(probas, labels, ignore=None):
    """Flattens predictions in the batch
    """
    if probas.dim() == 3:
        # assumes output of a sigmoid layer
        B, H, W = probas.size()
        probas = probas.view(B, 1, H, W)
    B, C, H, W = probas.size()
    probas = probas.permute(0, 2, 3, 1).contiguous().view(-1, C)  # B * H * W, C = P, C
    labels = labels.view(-1)
    if ignore is None:
        return probas, labels
    valid = labels != ignore
    vprobas = probas[valid.nonzero().squeeze()]
    vlabels = labels[valid]
    return vprobas, vlabels


# --------------------------- HELPER FUNCTIONS ---------------------------
def isnan(x):
    return x != x


def mean(values, ignore_nan=False, empty=0):
    """Nanmean compatible with generators.
    """
    values = iter(values)
    if ignore_nan:
        values = ifilterfalse(isnan, values)
    try:
        n = 1
        acc = next(values)
    except StopIteration:
        if empty == "raise":
            raise ValueError("Empty mean")
        return empty
    for n, v in enumerate(values, 2):
        acc += v
    if n == 1:
        return acc
    return acc / n


class BinaryLovaszLoss(_Loss):
    def __init__(self, per_image=False, ignore=None):
        super().__init__()
        self.ignore = ignore
        self.per_image = per_image

    def forward(self, logits, target):
        return _lovasz_hinge(logits, target, per_image=self.per_image, ignore=self.ignore)


class LovaszLoss(_Loss):
    def __init__(self, per_image=False, ignore=None):
        super().__init__()
        self.ignore = ignore
        self.per_image = per_image

    def forward(self, logits, target):
        return _lovasz_softmax(logits, target, per_image=self.per_image, ignore=self.ignore)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Dec 25 23:26:55 2020
@author: hasan
"""
import torch
import torch.nn as nn
import torch.nn.functional as F

'''
My own implementation of spatial relation and channel relation modules
Paper : A Relation-Augmented Fully Convolutional Network for Semantic Segmentation in Aerial Scenes
link : https://openaccess.thecvf.com/content_CVPR_2019/papers/Mou_A_Relation-Augmented_Fully_Convolutional_Network_for_Semantic_Segmentation_in_Aerial_CVPR_2019_paper.pdf
'''
class SR_Module(nn.Module):
    def __init__(self,
                 in_channels):
        
        super(SR_Module,self).__init__()
        
        self.conv1 = nn.Conv2d(in_channels = in_channels,
                               out_channels = in_channels,
                               kernel_size = 1,
                               padding = 0,
                               stride = 1)
        self.conv2 = nn.Conv2d(in_channels = in_channels,
                               out_channels = in_channels,
                               kernel_size = 1,
                               padding = 0,
                               stride = 1)
        self.relu = nn.ReLU()
        
    def forward(self,feats):
        B, C, H, W = feats.size()
        
        us = self.conv1(feats)
        vs = self.conv2(feats)
        us = us.view(B,C,-1) #(B, C, HxW)
        
        vs = vs.view(B,C,-1) #(B, C, HxW)
        us = us.permute(0,2,1) #(B, HxW, C)
        relations = torch.matmul(us,vs) #(B, HxW, HxW)
        relations = relations.view(B, H*W, H, W)
        relations = self.relu(relations)
        
        aug_feats = torch.cat((feats,relations),dim = 1)
        return aug_feats

class CR_Module(nn.Module):
    def __init__(self,
                 in_channels):
        super(CR_Module,self).__init__()
        self.gap = nn.AdaptiveAvgPool2d((1,1))
        self.conv1 = nn.Conv2d(in_channels = in_channels,
                               out_channels = in_channels,
                               kernel_size = 1,
                               padding = 0,
                               stride = 1)
        self.conv2 = nn.Conv2d(in_channels = in_channels,
                               out_channels = in_channels,
                               kernel_size = 1,
                               padding = 0,
                               stride = 1)
    def forward(self,feats):
        B, C, H, W = feats.size()
        descriptors = self.gap(feats)
        us = self.conv1(descriptors)
        vs = self.conv2(descriptors)
        us = us.view(B,C,-1)
        vs = vs.view(B,C,-1).permute(0,2,1)
        relations = torch.matmul(us, vs)
        relations = F.softmax(relations,dim=1)
        feats = feats.view(B,C,-1).permute(0,2,1)
        aug_feats = torch.matmul(feats, relations)
        aug_feats = aug_feats.view(B,C,H,W)
        return aug_feats
    
class S_RA_Module(nn.Sequential):
    def __init__(self,in_channels):
        
        super().__init__(CR_Module(in_channels),
                         SR_Module(in_channels))
        
class P_RA_Module(nn.Module):
    
    def __init__(self,in_channels):
        super(P_RA_Module,self).__init__()
        self.crm = CR_Module(in_channels)
        self.srm = SR_Module(in_channels)
        
    def forward(self,feats):
        rc = self.crm(feats)
        rs = self.srm(feats)
        relations = torch.cat((rs,rc),dim=1)
        return relations
    
def get_relational_module(name = 'S_RA',in_channels = 256, in_size = (16,16)):
    __mods__ = ['S_RA_M','P_RA_M','CR_M','SR_M']
    assert name in __mods__, f'{name} is not a valid relational module!'
    H,W = in_size
    if(name == 'CR_M'):
        m = CR_Module(in_channels)
        out_channels = in_channels
    elif(name == 'SR_M'):
        m = SR_Module(in_channels)
        out_channels = in_channels + H * W
    elif(name == 'S_RA_M'):
        m = S_RA_Module(in_channels)
        out_channels = in_channels + H * W
    elif(name == 'P_RA_M'):
        m = P_RA_Module(in_channels)
        out_channels = 2 * in_channels + H * W
    return m, out_channels
        
class Relational_Module(nn.Module):
    def __init__(self,
                 name,
                 in_channels,
                 in_size):
        super().__init__()
        self.mod ,self.out_channels = get_relational_module(name,in_channels,in_size)
        
    def forward(self,feats):
        aug_feats = self.mod(feats)
        return aug_feats
        
        
'''      
x= torch.rand(size=(1,128,16,16))
print(x.size())
m=P_RA_Module(x.size(1))

y = m(x)
print(y.size())
'''


import torch
from torch.nn.functional import interpolate, pad
import numpy as np
from typing import List


def remove_boundary_positives(img: torch.Tensor, pixels: int = 20) -> torch.Tensor:
    r"""Remove boundary pixels from input image tensor.

    Parameters
    ----------
    img : torch.Tensor
        Input image tensor with shape (B, C, H, W), where B is the batch size,
        C is the number of channels, H is the height, and W is the width.
    pixels : int, optional
        Number of pixels to remove from each boundary of the image. Default is 20.

    Returns
    -------
    torch.Tensor
        Output tensor with the same shape as the input image tensor,
        but with boundary pixels set to zero.
    """
    H, W = img.shape[-2:]
    bg = torch.zeros_like(img, dtype=img.dtype, device=img.device)

    s1 = min(pixels, H - 1)
    e1 = max(s1 + 1, H - pixels)
    s2 = min(pixels, W - 1)
    e2 = max(s2 + 1, W - pixels)

    bg[..., s1:e1, s2:e2] = img[..., s1:e1, s2:e2]
    return bg


def remove_boundary_positives_np(img: np.ndarray, pixels: int = 20) -> np.ndarray:
    r"""Remove boundary pixels from input image numpy array.

    Parameters
    ----------
    img : np.ndarray
        Input image numpy array with shape (H, W, C), where H is the height,
        W is the width, and C is the number of channels.
    pixels : int, optional
        Number of pixels to remove from each boundary of the image. Default is 20.

    Returns
    -------
    np.ndarray
        Output numpy array with the same shape as the input image array,
        but with boundary pixels set to zero.
    """
    H, W = img.shape[:2]
    bg = np.zeros_like(img, dtype=img.dtype)

    s1 = min(pixels, H - 1)
    e1 = max(s1 + 1, H - pixels)
    s2 = min(pixels, W - 1)
    e2 = max(s2 + 1, W - pixels)

    bg[s1:e1, s2:e2, ...] = img[s1:e1, s2:e2, ...]
    return bg


def resize_pad(x: torch.Tensor, padsize: int = None, resize: int = None, pad_value: int = -1) -> torch.Tensor:
    r"""Resize and/or pad input tensor.

    Parameters
    ----------
    x : torch.Tensor
        Input tensor to be resized and/or padded.
    padsize : int, optional
        int specifying the number of pixels to pad on each side of the input tensor. Default is None.
    resize : int , optional
        int specifying the target size of the output tensor. Default is None.
    pad_value : int, optional
        Value to use for padding. Default is -1.

    Returns
    -------
    torch.Tensor
        Output tensor with the specified resizing and/or padding applied.
    """
    if padsize is None and resize is None:
        return x
    
    input_shape = x.shape
    if len(input_shape) == 5:
        B,T,C,H,W = input_shape
        x = x.view(B*T,C,H,W)
        
    if resize is not None:
        x = interpolate(x,size=(resize,resize),mode='bilinear')

    if padsize is not None:
        if resize is not None:
            ppix = padsize - resize 
        else:
            ppix = padsize - 256
        s = ppix // 2
        e = ppix - s
        x = pad(x, (s,e,s,e), mode='constant', value=pad_value)

    
    if len(input_shape) == 5:
        H,W = x.shape[-2:]
        x = x.view(B,T,C,H,W)
    
    return x


def unpad_resize(x: torch.Tensor, padsize: int = None, resize: int = None) -> torch.Tensor:
    r"""Unpad and/or resize input tensor.

    Parameters
    ----------
    x : torch.Tensor
        Input tensor with shape (B, C, H, W), where B is the batch size,
        C is the number of channels, H is the height, and W is the width.
    padsize : int
        List specifying the pad size for each dimension [pad_top, pad_bottom, pad_left, pad_right].
    resize : int
        List specifying the target size for resizing [new_height, new_width].

    Returns
    -------
    torch.Tensor
        Output tensor with the specified unpadded and/or resized dimensions.
    """
    if padsize is None and resize is None:
        return x
    if padsize is not None:
        if resize is not None:
            ppix = padsize - resize 
        else:
            ppix = padsize - 256
        
        s = ppix // 2
        e = ppix - s
        H,W = x.shape[-2:]
        x = x[...,s:H-e,s:W-e]
        
    if resize is not None:
        x = interpolate(x,size=(256,256),mode='bilinear')

    return x




"""
This code is generated by Ridvan Salih KUZU @DLR
LAST EDITED:  14.09.2021
ABOUT SCRIPT:
It defines a data reader for Planet Fusion eath observation data
"""
import torch
import geopandas as gpd
import rasterio as rio
from rasterio import features
import numpy as np
import os
import zipfile
import glob
from tqdm import tqdm

class PlanetReader(torch.utils.data.Dataset):
    """
    THIS CLASS INITIALIZES THE DATA READER FOR PLANET DATA
    """
    def __init__(self, input_dir, label_dir, label_ids=None, transform=None, min_area_to_ignore = 1000,  selected_time_points=None):
        '''
        THIS FUNCTION INITIALIZES DATA READER.
        :param input_dir: directory of input images in TIF format
        :param label_dir: directory of ground-truth polygons in GeoJSON format
        :param label_ids: an array of crop IDs in order. if the crop labels in GeoJSON data is not started from index 0 it can be used. Otherwise it is not required.
        :param transform: data transformer function for the augmentation or data processing
        :param min_area_to_ignore: threshold m2 to eliminate small agricultural fields less than a certain threshold. By default, threshold is 1000 m2
        :param selected_time_points: If a sub set of the time series will be exploited, it can determine the index of those times in a given time series dataset

        :return: None
        '''

        self.data_transform = transform
        self.selected_time_points = selected_time_points
        self.crop_ids=label_ids
        if label_ids is not None and not isinstance(label_ids, list):
            self.crop_ids=label_ids.tolist()

        self.npyfolder = os.path.abspath(input_dir + "time_series")
        self.labels = PlanetReader._setup(input_dir, label_dir, self.npyfolder, min_area_to_ignore)


    def __len__(self):
        """
         THIS FUNCTION RETURNS THE LENGTH OF DATASET
         """
        return len(self.labels)

    def __getitem__(self, item):
        """
        THIS FUNCTION ITERATE OVER THE DATASET BY GIVEN ITEM NO AND RETURNS FOLLOWINGS:
        :return: image_stack in size of [Time Stamp, Image Dimension (Channel), Height, Width] , crop_label, field_mask in size of [Height, Width], field_id
        """

        feature = self.labels.iloc[item]

        npyfile = os.path.join(self.npyfolder, "fid_{}.npz".format(feature.fid))
        if os.path.exists(npyfile): # use saved numpy array if already created
            try:
                object = np.load(npyfile)
                image_stack = object["image_stack"]
                mask = object["mask"]
            except zipfile.BadZipFile:
                print("ERROR: {} is a bad zipfile...".format(npyfile))
                raise
        else:
            print("ERROR: {} is a missing...".format(npyfile))
            raise

        if self.data_transform is not None:
            image_stack, mask = self.data_transform(image_stack, mask)

        if self.selected_time_points is not None:
            image_stack = image_stack[self.selected_time_points]

        if self.crop_ids is not None:
            label = self.crop_ids.index(feature.crop_id)
        else:
            label = feature.crop_id

        return image_stack, label, mask, feature.fid


    @staticmethod
    def _setup(input_dir, label_dir, npyfolder, min_area_to_ignore=1000):
        """
        THIS FUNCTION PREPARES THE PLANET READER BY SPLITTING AND RASTERIZING EACH CROP FIELD AND SAVING INTO SEPERATE FILES FOR SPEED UP THE FURTHER USE OF DATA.
        :param input_dir: directory of input images in TIF format
        :param label_dir: directory of ground-truth polygons in GeoJSON format
        :param npyfolder: folder to save the field data for each field polygon
        :param min_area_to_ignore: threshold m2 to eliminate small agricultural fields less than a certain threshold. By default, threshold is 1000 m2
        :return: labels of the saved fields
        """

        inputs = glob.glob(input_dir + '/*/*.tif', recursive=True)
        tifs = sorted(inputs)
        labels = gpd.read_file(label_dir)

        # read coordinate system of tifs and project labels to the same coordinate reference system (crs)
        with rio.open(tifs[0]) as image:
            crs = image.crs
            print('INFO: Coordinate system of the data is: {}'.format(crs))
            transform = image.transform

        mask = labels.geometry.area > min_area_to_ignore
        print(f"INFO: Ignoring {(~mask).sum()}/{len(mask)} fields with area < {min_area_to_ignore}m2")

        labels = labels.loc[mask]
        labels = labels.to_crs(crs) #TODO: CHECK IF REQUIRED

        for index, feature in tqdm(labels.iterrows(), total=len(labels), position=0, leave=True, desc="INFO: Extracting time series into the folder: {}".format(npyfolder)):

            npyfile = os.path.join(npyfolder, "fid_{}.npz".format(feature.fid))
            if not os.path.exists(npyfile):

                left, bottom, right, top = feature.geometry.bounds
                window = rio.windows.from_bounds(left, bottom, right, top, transform)

                # reads each tif in tifs on the bounds of the feature. shape T x D x H x W
                image_stack = np.stack([rio.open(tif).read(window=window) for tif in tifs])

                with rio.open(tifs[0]) as src:
                    win_transform = src.window_transform(window)

                out_shape = image_stack[0, 0].shape
                assert out_shape[0] > 0 and out_shape[1] > 0, "WARNING: fid:{} image stack shape {} is zero in one dimension".format(feature.fid,image_stack.shape)

                # rasterize polygon to get positions of field within crop
                mask = features.rasterize(feature.geometry, all_touched=True,transform=win_transform, out_shape=image_stack[0, 0].shape)

                #mask[mask != feature.fid] = 0
                #mask[mask == feature.fid] = 1
                os.makedirs(npyfolder, exist_ok=True)
                np.savez(npyfile, image_stack=image_stack, mask=mask, feature=feature.drop("geometry").to_dict())


        return labels


if __name__ == '__main__':
    """
    EXAMPLE USAGE OF DATA READER
    """

    zippath = "../data/dlr_fusion_competition_germany_train_source_planet_5day"

    labelgeojson = "../data/dlr_fusion_competition_germany_train_labels/dlr_fusion_competition_germany_train_labels_33N_18E_242N/labels.geojson"
    ds = PlanetReader(zippath, labelgeojson, selected_time_points=[2,3,4])
    X,y,m,fid = ds[0]


import numpy as np
import torch
import matplotlib.pyplot as plt
import cv2
import geopandas as gpd
import os
import json
import glob
from tqdm import tqdm
import shapely.geometry as sg
from shapely import affinity
from shapely.geometry import Point, Polygon
import random
from PIL import Image, ImageDraw
from skimage import measure
import rasterio
from rasterio.features import geometry_mask
#from metrics import DiceScore,IoUScore
import pandas as pd
import torch.nn as nn
import argparse
from tqdm import trange,tqdm
import geopandas as gp
from shapely.geometry import Polygon
from rasterio.coords import BoundingBox
from rasterio import windows
from rasterio import warp
from rasterio.transform import from_bounds
from PIL import Image,ImageDraw
from skimage.morphology import dilation, square
from skimage.segmentation import watershed
from simplification.cutil import simplify_coords_vwp
from imantics import Mask
import numpy as np
import rasterio
from matplotlib import pyplot as plt
from rasterio.windows import Window
from typing import List, Tuple


def binary_mask_to_polygon(binary_mask):
    """
    @param binary_mask = binary mask to be converted in to polygon
    @type binary_mask = numpy array

    Returns:
    polygon = converted polygon
    """
  
    contours = measure.find_contours(binary_mask, 0.5)
    max_contour = max(contours, key=len)

    polygon = Polygon([(int(point[1]), int(point[0])) for point in max_contour])

    return polygon

def convert_polygon_to_mask(geo,shape,transform=None):
    """
    @param geo = polygons' geometry to be converted
    @param shape = shape of the mask to be generated
    @param transform = flag param if polygons are georeferenced
    @type geo = geopandas ['geometry']

    Returns:
    gtmask = mask of type array
    """

    gtmask=np.zeros(shape)
    if transform:
       for orig_row in geo:
          polygon=[]
          if orig_row.geom_type=="Polygon":
              binary_array = geometry_mask([orig_row], out_shape=shape, transform=transform, invert=True)
              ba=binary_array*1
              gtmask=gtmask+ba
          else:
              for x in orig_row.geoms:
                binary_array = geometry_mask([x], out_shape=shape, transform=transform, invert=True)
                ba=binary_array*1
                gtmask=gtmask+ba
    else:
      for orig_row in geo:
            polygon=[]
            if orig_row.geom_type=="Polygon":
                for point in orig_row.exterior.coords:
                    polygon.append(point)
                img = Image.new('L', shape, 0)
                ImageDraw.Draw(img).polygon(polygon, outline=1, fill=1)
                gt_mask_building = np.array(img)
                gtmask=gtmask+gt_mask_building
            else:
                for x in orig_row.geoms:
                  for point in x.exterior.coords:
                    polygon.append(point)

                img = Image.new('L', shape, 0)
                ImageDraw.Draw(img).polygon(polygon, outline=1, fill=1)
                gt_mask_building = np.array(img)
                gtmask=gtmask+gt_mask_building
    return gtmask


def convert_polygon_to_mask_batch(geo,shape,transform):
  """
    @param geo = polygons' geometry to be converted
    @param shape = shape of the mask to be generated
    @param transform = transformation information for polygons
    @type geo = geopandas ['geometry']

    Returns:
    gtmask = list of numpy array masks, each contain a polygon
    """ 
  gtmask=[]
  if transform:
     for row in geo:
      if row.geom_type=="Polygon":
        binary_array = geometry_mask([row], out_shape=shape, transform=transform, invert=True)
        ba=binary_array*1
        gtmask.append(ba)

      else:
        for x in row.geoms:
          binary_array = geometry_mask([x], out_shape=shape, transform=transform, invert=True)
          ba=binary_array*1
          gtmask.append(ba)
  else:   
    for orig_row in geo:
      polygon=[]
      if orig_row.geom_type=="Polygon":
          for point in orig_row.exterior.coords:
            polygon.append(point)
          img = Image.new('L', shape,0)
          ImageDraw.Draw(img).polygon(polygon, outline=1, fill=1)
          img=np.array(img)
          gtmask.append(img)
      else:
          for x in orig_row.geoms:
            for point in x.exterior.coords:
              polygon.append(point)
          img = Image.new('L', shape,0)
          ImageDraw.Draw(img).polygon(polygon, outline=1, fill=1)
          img=np.array(img)
          gtmask.append(img)
  
  return gtmask


def generate_polygon(bbox: List[float]) -> List[List[float]]:
    """
    Generates a list of coordinates forming a polygon.

    Parameters
    ----------
    bbox : List[float]
        A list representing the bounding box coordinates [xmin, ymin, xmax, ymax].

    Returns
    -------
    List[List[float]]
        A list of coordinates representing the polygon.
    """

    return [
        [bbox[0], bbox[1]],
        [bbox[2], bbox[1]],
        [bbox[2], bbox[3]],
        [bbox[0], bbox[3]],
        [bbox[0], bbox[1]]
    ]

def shape_polys(polyg: List[Polygon]) -> List[List[Tuple[float, float]]]:
    """Shapes the building polygons as a list of polygon lists.

    Parameters
    ----------
    polyg : List[Polygon]
        List of building polygons.

    Returns
    -------
    List[List[Tuple[float, float]]]
        List of shaped polygons.
    """

    all_polys = []
    for poly in polyg:
        if len(poly) >= 3:
            f = poly.reshape(-1, 2)
            simplified_vw = simplify_coords_vwp(f, .3)
            if len(simplified_vw) > 2:
                mpoly = []  
                for i in simplified_vw:
                    mpoly.append((i[0], i[1]))  
                mpoly.append((simplified_vw[0][0], simplified_vw[0][1]))
                all_polys.append(mpoly)
    return all_polys

def pol_to_np(pol: List[List[float]]) -> np.ndarray:
    """Converts a list of coordinates to a NumPy array.

    Parameters
    ----------
    pol : List[List[float]]
        List of coordinates: [[x1, y1], [x2, y2], ..., [xN, yN]].

    Returns
    -------
    np.ndarray
        NumPy array of coordinates.
    """
    
    return np.array([list(l) for l in pol])

def pol_to_bounding_box(pol: List[List[float]]) -> BoundingBox:
    """Converts a list of coordinates to a bounding box.

    Parameters
    ----------
    pol : List[List[float]]
        List of coordinates: [[x1, y1], [x2, y2], ..., [xN, yN]].

    Returns
    -------
    BoundingBox
        Bounding box of the coordinates.
    """

    arr = pol_to_np(pol)
    return BoundingBox(np.min(arr[:, 0]),
                       np.min(arr[:, 1]),
                       np.max(arr[:, 0]),
                       np.max(arr[:, 1]))



def reverse_coordinates(pol: List) -> List:
    """
    Reverse the coordinates in a polygon.

    Parameters
    ----------
    pol : list of list
        List of coordinates: [[x1, y1], [x2, y2], ..., [xN, yN]].

    Returns
    -------
    list of list
        Reversed coordinates: [[y1, x1], [y2, x2], ..., [yN, xN]].
    """

    return [list(f[-1::-1]) for f in pol]

class ArgMax(nn.Module):

    def __init__(self, dim=None):
        super().__init__()
        self.dim = dim

    def forward(self, x):
        return torch.argmax(x, dim=self.dim)
    



import torch
import unittest
from unittest import TestCase

class TestMetrics(TestCase):
    def test_dice_score(self):
        dice_score = DiceScore()

        # Define sample tensors
        y_pred = torch.tensor([[[[0.6, 0.4], [0.5, 0.2]], [[0.3, 0.8], [0.6, 0.9]]]])
        y_gt = torch.tensor([[[[1, 0], [1, 0]], [[0, 1], [0, 1]]]])

        # Compute Dice score
        dice_score_value = dice_score(y_pred, y_gt)
        expected_dice_score = torch.tensor(0.9)

        self.assertTrue(torch.allclose(dice_score_value, expected_dice_score, atol=1e-4))

    def test_iou_score(self):
        iou_score = IoUScore()

        y_pred = torch.tensor([[[[0.6, 0.4], [0.5, 0.2]], [[0.3, 0.8], [0.6, 0.9]]]])
        y_gt = torch.tensor([[[[1, 0], [1, 0]], [[0, 1], [0, 1]]]])

        iou_score_value = iou_score(y_pred, y_gt)
        expected_iou_score = torch.tensor(0.83)

        self.assertTrue(torch.allclose(iou_score_value, expected_iou_score, atol=1e-2))
        
    def test_threshold(self):
        input_tensor = torch.tensor([[0.6, 0.4], [0.3, 0.8]])
        threshold = 0.5

        # Apply thresholding
        thresholded_tensor = _threshold(input_tensor, threshold)
        expected_tensor = torch.tensor([[1, 0], [0, 1]])

        # Check if the actual output matches the expected output
        self.assertTrue(torch.all(torch.eq(thresholded_tensor, expected_tensor)))
